#include "Aaron.h"

BDebugTraceFunction BTestDebug;
#define EPSILON ((double)0.0088564516790356308171716757554635)
#define KAPPA   ((double)903.29629629629629629629629629629630)
#define EPSILONKAPPA ((double)8.0)

int MainLoopConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg  = BTileImage(pipeData->m_tile);

	// If we don't have an image, we can't do anything
	if (bimg == 0)
	{
		BMSG( "MainLoopConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);
	rw = BImageRowWords(bimg);
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);

	if ( nChan != 3 )
	{
		BMSG("MainLoopConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}
        
        // grab our settings from the widget
        int refwhite,dump;
        double xmul,ymul;
        double white_x,white_z;
        double white_x_out,white_z_out;
        double const white_y = 1.0;
        double const white_y_out = 1.0;
        bool adaptenable,sync;
        double matrix[9];
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          AARON_REFWHITE, bibbleData->m_plugin, refwhite);
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          AARON_DUMP, bibbleData->m_plugin, dump);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          AARON_XMUL, bibbleData->m_plugin, xmul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          AARON_YMUL, bibbleData->m_plugin, ymul);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          AARON_ADAPTENABLE, bibbleData->m_plugin, adaptenable);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          AARON_SYNC, bibbleData->m_plugin, sync);
        if(sync) { xmul = (xmul+ymul)/((double)2.0); ymul = xmul; }

        // http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html
        switch(refwhite) {
          case 0: { // A
            /*
              Average incandescent light
                http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.1
              CIE standard illuminant A is intended to represent typical, domestic, tungsten-filament lighting.  (source: idem)

              This represents an incandescent light source like a tungsten filament.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html

                Incandescent bulb simulator
              http://en.wikipedia.org/wiki/D65
            */
            white_x = (double) 1.09850;
            white_z = (double) 0.35585;
            break;
          }
          case 1: { // B
            /*
              noon sunlight
                http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.2

              Daylight simulator (direct)
              http://en.wikipedia.org/wiki/D65
            */
            white_x = (double) 0.99072;
            white_z = (double) 0.85223;
            break;
          }
          case 2: { // C
            /*
              day light
                http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.2

              Represents average daylight.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html

              Daylight simulator (shade)
              http://en.wikipedia.org/wiki/D65
            */
            white_x = (double) 0.98074;
            white_z = (double) 1.18232;
            break;
          }
          /*
            Illuminant series D
              "natural daylight"
              http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.3

              The D-series illuminants represent different times of the day in an outdoor setting.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html
          */
          case 3: { // D50
            white_x = (double) 0.96422;
            white_z = (double) 0.82521;
            break;
          }
          case 4: { // D55
            white_x = (double) 0.95682;
            white_z = (double) 0.92149;
            break;
          }
          case 5: { // D65
            white_x = (double) 0.95047;
            white_z = (double) 1.08883;
            break;
          }
          case 6: { // D75
            white_x = (double) 0.94972;
            white_z = (double) 1.22638;
            break;
          }
          case 7: { // E
            /*
              equal-energy radiator
                http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.4
            */
            white_x = (double) 1.00000;
            white_z = (double) 1.00000;
            break;
          }
          /*
            fluorescent lighting
              http://efi-vutek.ru/upload_files/Standardilluminant.pdf p.5

              These are the fluorescent light simulators.
              F2, F7, or F11 Illuminants - These illuminants simulate most indoor lighting.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html
          */
          case 8: { // F2
          /*
            F2 - Represents a typically cool white fluorescent source.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html
          */
            white_x = (double) 0.99186;
            white_z = (double) 0.67393;
            break;
          }
          case 9: { // F7
          /*
            F7 - Represents a broad band daylight fluorescent source.
              http://www.onyxgfx.net/Wiki/ICC%20Build%20Options.html

              broadband(F7)
              http://herkules.oulu.fi/isbn9514267885/html/i1030449.html
          */
            white_x = (double) 0.95041;
            white_z = (double) 1.08747;
            break;
          }
          case 10: { // F11
            /*
                three-band
                http://herkules.oulu.fi/isbn9514267885/html/i1030449.html
            */
            white_x = (double) 1.00962;
            white_z = (double) 0.64350;
            break;
          }
        }
        white_x_out = white_x;
        white_z_out = white_z;

        if(adaptenable) {
          int targetwhite;
          bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
            AARON_TARGETWHITE, bibbleData->m_plugin, targetwhite);
          if(refwhite==targetwhite) {
            adaptenable = false;
          } else {
            int adaptmethod;
            bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
              AARON_ADAPTMETHOD, bibbleData->m_plugin, adaptmethod);
#include "adapt.c"
            switch(targetwhite) {
              case 0: { // A
                white_x_out = (double) 1.09850;
                white_z_out = (double) 0.35585;
                break;
              }
              case 1: { // B
                white_x_out = (double) 0.99072;
                white_z_out = (double) 0.85223;
                break;
              }
              case 2: { // C
                white_x_out = (double) 0.98074;
                white_z_out = (double) 1.18232;
                break;
              }
              case 3: { // D50
                white_x_out = (double) 0.96422;
                white_z_out = (double) 0.82521;
                break;
              }
              case 4: { // D55
                white_x_out = (double) 0.95682;
                white_z_out = (double) 0.92149;
                break;
              }
              case 5: { // D65
                white_x_out = (double) 0.95047;
                white_z_out = (double) 1.08883;
                break;
              }
              case 6: { // D75
                white_x_out = (double) 0.94972;
                white_z_out = (double) 1.22638;
                break;
              }
              case 7: { // E
                white_x_out = (double) 1.00000;
                white_z_out = (double) 1.00000;
                break;
              }
              case 8: { // F2
                white_x_out = (double) 0.99186;
                white_z_out = (double) 0.67393;
                break;
              }
              case 9: { // F7
                white_x_out = (double) 0.95041;
                white_z_out = (double) 1.08747;
                break;
              }
              case 10: { // F11
                white_x_out = (double) 1.00962;
                white_z_out = (double) 0.64350;
                break;
              }
            }
          }
        }

     // double white_little_x = white_x/(white_x+white_y+white_z);
     // double white_little_y = white_y/(white_x+white_y+white_z);
        double white_little_x_out =
          white_x_out/(white_x_out+white_y_out+white_z_out);
        double white_little_y_out =
          white_y_out/(white_x_out+white_y_out+white_z_out);

        if(dump==2) { // Initial Y
          xmul = ymul = (double) 0.0;
          adaptenable = false;
        }
        if(dump==3) { // Final Y
          xmul = ymul = (double) 0.0;
        }

	unsigned short *pIn, *pEnd;
        double R,G,B;

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {

if(dump==1) { // Initial L
  R = ((double) *(pIn));
  G = B = (double)32768.0;
} else {
#include "core.c"
  if(dump==4) { // Final L
    G = B = (double)32768.0;
  }
}
				*(pIn) = (unsigned short)(R + 0.5f);
				*(pIn + pw) = (unsigned short)(G + 0.5f);
				*(pIn + 2*pw) = (unsigned short)(B + 0.5f);
				pIn += np;
			}
		}
	} else {
	        unsigned short *pAlpha;
	        BImageHandle alpha = pipeData->m_alpha; 
		float ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pAlpha += np;
					continue;
				}
if(dump==1) { // Initial L
  R = ((double) *(pIn));
  G = B = 32768;
} else {
#include "core.c"
  if(dump==4) { // Final L
    G = B = 32768;
  }
}
				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pIn) = (unsigned short)(R + 0.5f);
					*(pIn + pw) = (unsigned short)(G + 0.5f);
					*(pIn + 2*pw) = (unsigned short)(B + 0.5f);
				} else {
					oval = pIn[0]*ival + R*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[0] = oval;

					oval = pIn[pw]*ival + G*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw] = oval;

					oval = pIn[pw+pw]*ival + B*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw+pw] = oval;

				}
				pIn += np; pAlpha += np;
			}
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	BComboBoxHandle cb;
	cb = BSettingsWidgetGetCombo( sw, "aaron_refwhite_combo" );
	if (cb==0)
	{
		BMSG("Aaron Plugin couldn't find aaron_refwhite ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "A (incandescent)", 0 );
	BComboBoxAppend(cb, "B (noon sunlight)", 1 );
	BComboBoxAppend(cb, "C (day shade)", 2 );
	BComboBoxAppend(cb, "D50 (natural)", 3 );
	BComboBoxAppend(cb, "D55 (natural)", 4 );
	BComboBoxAppend(cb, "D65 (natural)", 5 );
	BComboBoxAppend(cb, "D75 (natural)", 6 );
	BComboBoxAppend(cb, "E (equal-energy)", 7 );
	BComboBoxAppend(cb, "F2 (fluorescent)", 8 );
	BComboBoxAppend(cb, "F7 (fluorescent)", 9 );
	BComboBoxAppend(cb, "F11 (fluorescent)", 10 );

        // Default
	BComboBoxSetPos(cb, 3);




	cb = BSettingsWidgetGetCombo( sw, "aaron_targetwhite_combo" );
	if (cb==0)
	{
		BMSG("Aaron Plugin couldn't find aaron_targetwhite ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "A (incandescent)", 0 );
	BComboBoxAppend(cb, "B (noon sunlight)", 1 );
	BComboBoxAppend(cb, "C (day shade)", 2 );
	BComboBoxAppend(cb, "D50 (natural)", 3 );
	BComboBoxAppend(cb, "D55 (natural)", 4 );
	BComboBoxAppend(cb, "D65 (natural)", 5 );
	BComboBoxAppend(cb, "D75 (natural)", 6 );
	BComboBoxAppend(cb, "E (equal-energy)", 7 );
	BComboBoxAppend(cb, "F2 (fluorescent)", 8 );
	BComboBoxAppend(cb, "F7 (fluorescent)", 9 );
	BComboBoxAppend(cb, "F11 (fluorescent)", 10 );

        // Default
	BComboBoxSetPos(cb, 3);


	cb = BSettingsWidgetGetCombo( sw, "aaron_adaptmethod_combo" );
	if (cb==0)
	{
		BMSG("Aaron Plugin couldn't find aaron_adaptmethod ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "XYZ scaling", 0 );
	BComboBoxAppend(cb, "Bradford method", 1 );
	BComboBoxAppend(cb, "Von Kries", 2 );

        // Default
	BComboBoxSetPos(cb, 1);



	cb = BSettingsWidgetGetCombo( sw, "aaron_dump_combo" );
	if (cb==0)
	{
		BMSG("Aaron Plugin couldn't find aaron_dump ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "(color output)", 0 );
	BComboBoxAppend(cb, "Dump initial L", 1 );
	BComboBoxAppend(cb, "Dump initial Y", 2 );
	BComboBoxAppend(cb, "Dump final Y", 3 );
	BComboBoxAppend(cb, "Dump final L", 4 );

        // Default
	BComboBoxSetPos(cb, 0);

	return rc;
}

/*
  The following function is called in order to fix dependancies between
  settings (for instance, if some user has to choose a color with three
  sliders (red, green, blue), you may want to change sliders for some
  Hue, saturation, value model of the same color (HSV) in order to
  allow the user to choose a color by two different (but linked) ways.
  You may see some dependancies with the Simple Color Filter v1.1 (just
  set the color with the sliders, and see the color of the button changing
  in the same time).
  The most known feature handled here is enabling the plugin when the user
  moves a slider.
*/
int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Aaron - Do Dependancies");
#endif

        // A nice feature: when the user will change some settings,
        // the plugin will get enabled "automagically"
        // The double-bar || means "or": if there is a change on
        // some option OR on some other option, then enable the plugin.
	if ((BOptionListContains(od->m_changeList,AARON_REFWHITE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_XMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_YMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_DUMP,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_ADAPTENABLE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_TARGETWHITE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_ADAPTMETHOD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_SYNC,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,AARON_ENABLED,bibbleData->m_plugin,true);
	}
	if ((BOptionListContains(od->m_changeList,AARON_TARGETWHITE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,AARON_ADAPTMETHOD,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,AARON_ADAPTENABLE,bibbleData->m_plugin,true);
	}

	if ( BOptionListContains( od->m_changeList, AARON_REFWHITE, bibbleData->m_plugin) )
	{
	  int refwhite;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "aaron_refwhite_combo" );	
		BGetOptionInt( od->m_changeList, AARON_REFWHITE, bibbleData->m_plugin, refwhite);
		BComboBoxSetPos(cb, refwhite);
	}
	if ( BOptionListContains( od->m_changeList, AARON_DUMP, bibbleData->m_plugin) )
	{
	  int dump;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "aaron_dump_combo" );	
		BGetOptionInt( od->m_changeList, AARON_DUMP, bibbleData->m_plugin, dump);
		BComboBoxSetPos(cb, dump);
	}
	if ( BOptionListContains( od->m_changeList, AARON_TARGETWHITE, bibbleData->m_plugin) )
	{
	  int targetwhite;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "aaron_targetwhite_combo" );	
		BGetOptionInt( od->m_changeList, AARON_TARGETWHITE, bibbleData->m_plugin, targetwhite);
		BComboBoxSetPos(cb, targetwhite);
	}
	if ( BOptionListContains( od->m_changeList, AARON_ADAPTMETHOD, bibbleData->m_plugin) )
	{
	  int adaptmethod;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "aaron_adaptmethod_combo" );	
		BGetOptionInt( od->m_changeList, AARON_ADAPTMETHOD, bibbleData->m_plugin, adaptmethod);
		BComboBoxSetPos(cb, adaptmethod);
	}

	return rc;
}

// This function is a registered button callback
int aaronHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Aaron - aaronHandleButton()" );
#endif

	BMsgBox( "Aaron", "Version 1.0b\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_groupName, AARON_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_pluginName, AARON_PLUGINNAME);

        // TUTORIAL: choose the stage where you want the plugin
        // to run (you may find the list in the BibblePlugin.h
        // file from the SDK); you may also choose some priority
        // (in order tomake it run before or after another one
        // in the same stage): high priority means that the
        // plugin will run earlier.
        // In order to make some tests, see if your plugin runs
        // before or after the Black and White plugin which has
        // the following settings: PIPE_PLUGIN_CURVES, 200
        // Try a lower and then a higher priority and look at
        // the result.
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, MainLoopConvert, PIPE_PLUGIN_COLORMANAGE2, 250);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, aaronHandleButton, AARON_BUTTONNAME);

        // The function registered below (doDependancies) will be called
        // when a setting will be changed
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies );

	BMSG("Aaron plugin loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

/*
  TUTORIAL:
  The following function is very easy to handle but you have to
  spend a little time on it. Add one line for each setting by
  copying or changing existing lines.
	baf->m_addIntOption    (for an integer value) 
	baf->m_addDoubleOption (for a double value)
	baf->m_addBoolOption   (for a boolean value, most often checkboxes)
  The third parameter is the name of the setting in the UI (without the
  suffix: "_slider", "_edit", "_checkbox", etc.
  The fourth parameter is the name of the setting in the "selective settings"
  panel (something human-readable but rather short: three or four words).
  The sixth parameter is the tooltip for this setting (but don't forget to
  copy the same tooltip in the UI file).
  The seventh parameter is the default value for this setting (but don't
  forget to put this default value in the UI which can be relevant in some
  situations).
  The eightth parameter tells Bibble how to handle this setting. You may choose
  several options (all are in the Bibble.h file from the SDK); if you want
  to use several, use a single vertical bar | between them (you may see
  an example in the first line below).
  Most often you can only put OPTION_RESTART which tells that Bibble has to
  restart the computation for this picture when the setting has changed.
  For the "enabled" checkbox I also added a second option to tell Bibble
  it does not need to check dependancies (see the doDependancies function
  above) when the user changes this setting: no other setting has to be
  linked to this action.
*/
__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, AARON_ENABLED, "aaron_enabled", "Enable Aaron", AARON_GROUP, "Enable Aaron.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
	baf->m_addIntOption( bibbleData->m_plugin, AARON_REFWHITE, "aaron_refwhite", "Reference illuminant", AARON_GROUP, "Set the reference white.\n\nThe standard illuminants are defined as follow:\n\nCIE standard illuminant A is intended to represent typical,\ndomestic, tungsten-filament lighting.\n\nIlluminants B and C are daylight simulators. B served as\na representative of noon sunlight, while C represented\naverage day light.\n\nThe D series of illuminants are constructed to represent\nnatural daylight.\nD50: horizon light;\nD55: mid-morning / mid-afternoon daylight;\nD65: noon daylight;\nD75: North sky daylight.\n\nIlluminant E is an equal-energy radiator.\n\nThe F series of illuminants represent various types of\nfluorescent lighting.\nF2: cool white fluorescent;\nF7: daylight simulator (broadband fluorescent lamp);\nF11: narrow triband fluorescent.", 3, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, AARON_DUMP, "aaron_dump", "Black & White", AARON_GROUP, "Set a grayscale output.", false, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, AARON_XMUL, "aaron_xmul", "Change x saturation", AARON_GROUP, "Change the saturation for x.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, AARON_YMUL, "aaron_ymul", "Change y saturation", AARON_GROUP, "Change the saturation for y.", 1.0, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, AARON_ADAPTENABLE, "aaron_adaptenable", "Enable chromatic adaptation", AARON_GROUP, "Enable chromatic adaptation.", false, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, AARON_TARGETWHITE, "aaron_targetwhite", "Target illuminant", AARON_GROUP, "Set the target illuminant.\n\nThe standard illuminants are defined as follow:\n\nCIE standard illuminant A is intended to represent typical,\ndomestic, tungsten-filament lighting.\n\nIlluminants B and C are daylight simulators. B served as\na representative of noon sunlight, while C represented\naverage day light.\n\nThe D series of illuminants are constructed to represent\nnatural daylight.\nD50: horizon light;\nD55: mid-morning / mid-afternoon daylight;\nD65: noon daylight;\nD75: North sky daylight.\n\nIlluminant E is an equal-energy radiator.\n\nThe F series of illuminants represent various types of\nfluorescent lighting.\nF2: cool white fluorescent;\nF7: daylight simulator (broadband fluorescent lamp);\nF11: narrow triband fluorescent.", 3, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, AARON_ADAPTMETHOD, "aaron_adaptmethod", "CA method", AARON_GROUP, "Set the method for performing chromatic adaptation.", 1, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, AARON_SYNC, "aaron_sync", "Synchronize sliders", AARON_GROUP, "Synchronize both sliders by taking the average value.", false, OPTION_RESTART);

	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, AARON_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
