#include "Bathsheba.h"

BDebugTraceFunction BTestDebug;
#define EPSILON ((double)0.0088564516790356308171716757554635)
#define KAPPA   ((double)903.29629629629629629629629629629630)
#define EPSILONKAPPA ((double)8.0)

int MainLoopConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg  = BTileImage(pipeData->m_tile);

	// If we don't have an image, we can't do anything
	if (bimg == 0)
	{
		BMSG( "MainLoopConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);
	rw = BImageRowWords(bimg);
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);

	if ( nChan != 3 )
	{
		BMSG("MainLoopConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}
        
        // grab our settings from the widget
        int refwhite,spectrum;
        double white_x,white_z;
        double const white_y = 1.0;
        double normalize;
        bool ychannel;
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          BATHSHEBA_REFWHITE, bibbleData->m_plugin, refwhite);
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          BATHSHEBA_SPECTRUM, bibbleData->m_plugin, spectrum);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          BATHSHEBA_YCHANNEL, bibbleData->m_plugin, ychannel);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          BATHSHEBA_NORMALIZE, bibbleData->m_plugin, normalize);
        double film[44];
        for(int i=0;i<44;i++) {
          bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
            indices[i], bibbleData->m_plugin, film[i]);
        }
#include "tables.h"
        double x0 = 0;
        double y0 = 0;
        double z0 = 0;
        double normalize_film;
        switch(spectrum) {
          case 0: { // Cie 1931
            for(int i=0;i<44;i++) {
              x0 += film[i]*xyz1x[i];
              y0 += film[i]*xyz1y[i];
              z0 += film[i]*xyz1z[i];
            }
            // TODO: look for the best value (left)
            normalize_film = normalize*((double)150.0);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
          case 1: { // Cie 1931-1951
            for(int i=0;i<44;i++) {
              x0 += film[i]*xyz2x[i];
              y0 += film[i]*xyz2y[i];
              z0 += film[i]*xyz2z[i];
            }
            // TODO: look for the best value (left)
            normalize_film = normalize*((double)150.0);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
          case 2: { // Cie 1931-1978
            for(int i=0;i<44;i++) {
              x0 += film[i]*xyz3x[i];
              y0 += film[i]*xyz3y[i];
              z0 += film[i]*xyz3z[i];
            }
            // TODO: look for the best value (left)
            normalize_film = normalize*((double)150.0);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
          case 3: { // Cie 1964
            for(int i=0;i<44;i++) {
              x0 += film[i]*xyz4x[i];
              y0 += film[i]*xyz4y[i];
              z0 += film[i]*xyz4z[i];
            }
            // TODO: look for the best value (left)
            normalize_film = normalize*((double)150.0);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
          case 4: { // D-illuminany
            for(int i=0;i<44;i++) {
              x0 += film[i]*dilluminants0[i];
              y0 += film[i]*dilluminants1[i];
              z0 += film[i]*dilluminants2[i];
            }
            // TODO: look for the best value (left)
            normalize_film = normalize/((double)4.0);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
          case 5: { // Bathsheba
            // TODO: look for the best value (left)
            normalize_film = normalize*((double)2.0e-12);
            if(ychannel) normalize_film *= (double)260.0;
            break;
          }
        }


        switch(refwhite) {
          case 0: { // A
            white_x = (double) 1.09850;
            white_z = (double) 0.35585;
            break;
          }
          case 1: { // B
            white_x = (double) 0.99072;
            white_z = (double) 0.85223;
            break;
          }
          case 2: { // C
            white_x = (double) 0.98074;
            white_z = (double) 1.18232;
            break;
          }
          case 3: { // D50
            white_x = (double) 0.96422;
            white_z = (double) 0.82521;
            break;
          }
          case 4: { // D55
            white_x = (double) 0.95682;
            white_z = (double) 0.92149;
            break;
          }
          case 5: { // D65
            white_x = (double) 0.95047;
            white_z = (double) 1.08883;
            break;
          }
          case 6: { // D75
            white_x = (double) 0.94972;
            white_z = (double) 1.22638;
            break;
          }
          case 7: { // E
            white_x = (double) 1.00000;
            white_z = (double) 1.00000;
            break;
          }
          case 8: { // F2
            white_x = (double) 0.99186;
            white_z = (double) 0.67393;
            break;
          }
          case 9: { // F7
            white_x = (double) 0.95041;
            white_z = (double) 1.08747;
            break;
          }
          case 10: { // F11
            white_x = (double) 1.00962;
            white_z = (double) 0.64350;
            break;
          }
        }

        //double white_little_x = white_x/(white_x+white_y+white_z);
        //double white_little_y = white_y/(white_x+white_y+white_z);

	// Convert each scan line including the padding to monochrome
	// May not be necessary to convert the pad, but any processing
        // after us that dips into the pad
	// may cause color artifacts around the edges of the image
	unsigned short *pIn, *pEnd, *pAlpha;
        double R,G,B;

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
#include "core.c"
				*(pIn) = (unsigned short)(R + 0.5f);
				*(pIn + pw) = (unsigned short)32768;
				*(pIn + 2*pw) = (unsigned short)32768;
				pIn += np;
			}
		}
	} else {
	        BImageHandle alpha = pipeData->m_alpha; 
		float ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pAlpha += np;
					continue;
				}
#include "core.c"
				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pIn) = (unsigned short)(R + 0.5f);
					*(pIn + pw) = (unsigned short)32768;
					*(pIn + 2*pw) = (unsigned short)32768;
				} else {
					oval = pIn[0]*ival + R*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[0] = oval;

					oval = pIn[pw]*ival + ((double)32768.0)*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw] = oval;

					oval = pIn[pw+pw]*ival + ((double)32768.0)*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw+pw] = oval;

				}
				pIn += np; pAlpha += np;
			}
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	BComboBoxHandle cb;
	cb = BSettingsWidgetGetCombo( sw, "bathsheba_refwhite_combo" );
	if (cb==0)
	{
		BMSG("Bathsheba Plugin couldn't find bathsheba_refwhite ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "A (incandescent)", 0 );
	BComboBoxAppend(cb, "B (noon sunlight)", 1 );
	BComboBoxAppend(cb, "C (day shade)", 2 );
	BComboBoxAppend(cb, "D50 (natural)", 3 );
	BComboBoxAppend(cb, "D55 (natural)", 4 );
	BComboBoxAppend(cb, "D65 (natural)", 5 );
	BComboBoxAppend(cb, "D75 (natural)", 6 );
	BComboBoxAppend(cb, "E (equal-energy)", 7 );
	BComboBoxAppend(cb, "F2 (fluorescent)", 8 );
	BComboBoxAppend(cb, "F7 (fluorescent)", 9 );
	BComboBoxAppend(cb, "F11 (fluorescent)", 10 );

        // Default
	BComboBoxSetPos(cb, 3);


	cb = BSettingsWidgetGetCombo( sw, "bathsheba_spectrum_combo" );
	if (cb==0)
	{
		BMSG("Bathsheba Plugin couldn't find bathsheba_spectrum ComboBox");
		return -1;
	}

	// A value of -1 means prepend
	BComboBoxClear(cb);

	BComboBoxAppend(cb, "XYZ (1931)", 0 );
	BComboBoxAppend(cb, "XYZ (1931-1951)", 1 );
	BComboBoxAppend(cb, "XYZ (1931-1978)", 2 );
	BComboBoxAppend(cb, "XYZ (1964)", 3 );
	BComboBoxAppend(cb, "D-illuminant", 4 );
	BComboBoxAppend(cb, "Blackbody", 5 );

        // Default
	BComboBoxSetPos(cb, 4);





  unsigned int tm[44] = {
    72,72,65,59,56,52,50,49,47,44,42,41,40,41,41,42,42,40,38,38,38,
    36,34,32,30,29,29,31,31,38,49,62,72,72,72,72,72,72,72,72,72,72,
    72,72
  };
  char label[400];
  for(int i=0;i<44;i++) {
    sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"#bbbbbb\"><tbody><tr><td> </td></tr></tbody></table></body></html>",tm[i]);
	bibbleData->m_bsw.m_setLabelText(sw,vbar[i],label);
  }

	return rc;
}

/*
  The following function is called in order to fix dependancies between
  settings (for instance, if some user has to choose a color with three
  sliders (red, green, blue), you may want to change sliders for some
  Hue, saturation, value model of the same color (HSV) in order to
  allow the user to choose a color by two different (but linked) ways.
  You may see some dependancies with the Simple Color Filter v1.1 (just
  set the color with the sliders, and see the color of the button changing
  in the same time).
  The most known feature handled here is enabling the plugin when the user
  moves a slider.
*/
int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Bathsheba - Do Dependancies");
#endif

	if ((BOptionListContains(od->m_changeList,BATHSHEBA_REFWHITE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_SPECTRUM,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_NORMALIZE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_YCHANNEL,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,BATHSHEBA_ENABLED,bibbleData->m_plugin,true);
	}

	if ( BOptionListContains( od->m_changeList, BATHSHEBA_REFWHITE, bibbleData->m_plugin) )
	{
	  int refwhite;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "bathsheba_refwhite_combo" );	
		BGetOptionInt( od->m_changeList, BATHSHEBA_REFWHITE, bibbleData->m_plugin, refwhite);
		BComboBoxSetPos(cb, refwhite);
	}
	if ( BOptionListContains( od->m_changeList, BATHSHEBA_SPECTRUM, bibbleData->m_plugin) )
	{
	  int spectrum;
	  BSettingsWidgetHandle wh = od->m_widget;
	  BComboBoxHandle cb = BSettingsWidgetGetCombo( wh, "bathsheba_spectrum_combo" );	
		BGetOptionInt( od->m_changeList, BATHSHEBA_SPECTRUM, bibbleData->m_plugin, spectrum);
		BComboBoxSetPos(cb, spectrum);
	}

if (BOptionListContains(od->m_changeList, BATHSHEBA_ADJUSTCURRENTWL, bibbleData->m_plugin)) {
  BSettingsWidgetHandle wh = od->m_widget;
  int index;
  BGetOptionInt( od->m_currentList, BATHSHEBA_SELECTCURRENTWL, bibbleData->m_plugin, index);
  double value;
  BGetOptionDouble( od->m_changeList, BATHSHEBA_ADJUSTCURRENTWL, bibbleData->m_plugin, value);
  BSetOptionDouble(od->m_changeList,indices[index],bibbleData->m_plugin,value);
  char label[400];
  sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"white\"><tbody><tr><td> </td></tr></tbody></table></body></html>",72-((int) (36*value + .5)));
  BSettingsWidgetSetLabel( wh, vbar[index],label);
}

if (BOptionListContains(od->m_changeList, BATHSHEBA_SELECTCURRENTWL, bibbleData->m_plugin)) {
  BSettingsWidgetHandle wh = od->m_widget;
  int index;
  BGetOptionInt( od->m_changeList, BATHSHEBA_SELECTCURRENTWL, bibbleData->m_plugin, index);
  int j;
  char label[400];
  sprintf(label,"%dnm",350+index*10);
  BSettingsWidgetSetLabel( wh, "bathsheba_adjustcurrentwl_label",label);
  for(j=0;j<44;j++) {
    double v;
    BGetOptionDouble(od->m_currentList,indices[j], bibbleData->m_plugin, v);
    if(j==index) {
          sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"white\"><tbody><tr><td> </td></tr></tbody></table></body></html>",72-((int) (36*v + .5)));
  // put the adjust slider at the right position
  BSetOptionDouble(od->m_changeList,BATHSHEBA_ADJUSTCURRENTWL,bibbleData->m_plugin,v);
    } else {
          sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"#bbbbbb\"><tbody><tr><td> </td></tr></tbody></table></body></html>",72-((int) (36*v + .5)));
    }
      BSettingsWidgetSetLabel( wh, vbar[j],label);
  }
}

	return rc;
}

// This function is a registered button callback
int bathshebaHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Bathsheba - bathshebaHandleButton()" );
#endif

	BMsgBox( "Bathsheba", "Version 2.0\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}

int hotnessChanged(struct BibblePluginData *bibbleData,struct OptionData *od)
{
	int rc =0;
#ifdef _DEBUG
	BMSG( "Bathsheba Hotness Changed");
#endif

  BSettingsWidgetHandle wh = od->m_widget;
  int index;
  BGetOptionInt( od->m_currentList, BATHSHEBA_SELECTCURRENTWL, bibbleData->m_plugin, index);
  int j;
  char label[400];
  sprintf(label,"%dnm",350+index*10);
  BSettingsWidgetSetLabel( wh, "bathsheba_adjustcurrentwl_label",label);
  for(j=0;j<44;j++) {
    double v;
    BGetOptionDouble(od->m_currentList,indices[j], bibbleData->m_plugin, v);
    if(j==index) {
          sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"white\"><tbody><tr><td> </td></tr></tbody></table></body></html>",72-((int) (36*v + .5)));
  // put the adjust slider at the right position
  BSetOptionDouble(od->m_changeList,BATHSHEBA_ADJUSTCURRENTWL,bibbleData->m_plugin,v);
    } else {
          sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 144px\"><table style=\"margin-top: %dpx\" width=\"4px\" bgcolor=\"#bbbbbb\"><tbody><tr><td> </td></tr></tbody></table></body></html>",72-((int) (36*v + .5)));
    }
      BSettingsWidgetSetLabel( wh, vbar[j],label);
  }
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_groupName, BATHSHEBA_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_pluginName, BATHSHEBA_PLUGINNAME);

        // TUTORIAL: choose the stage where you want the plugin
        // to run (you may find the list in the BibblePlugin.h
        // file from the SDK); you may also choose some priority
        // (in order tomake it run before or after another one
        // in the same stage): high priority means that the
        // plugin will run earlier.
        // In order to make some tests, see if your plugin runs
        // before or after the Black and White plugin which has
        // the following settings: PIPE_PLUGIN_CURVES, 200
        // Try a lower and then a higher priority and look at
        // the result.
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, MainLoopConvert, PIPE_PLUGIN_COLORMANAGE2, 225);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, bathshebaHandleButton, BATHSHEBA_BUTTONNAME);

        // The function registered below (doDependancies) will be called
        // when a setting will be changed
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies );
	bibbleData->m_brf.m_registerHotnessChanged(bibbleData->m_plugin, hotnessChanged);

	BMSG("Bathsheba plugin loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

/*
  TUTORIAL:
  The following function is very easy to handle but you have to
  spend a little time on it. Add one line for each setting by
  copying or changing existing lines.
	baf->m_addIntOption    (for an integer value) 
	baf->m_addDoubleOption (for a double value)
	baf->m_addBoolOption   (for a boolean value, most often checkboxes)
  The third parameter is the name of the setting in the UI (without the
  suffix: "_slider", "_edit", "_checkbox", etc.
  The fourth parameter is the name of the setting in the "selective settings"
  panel (something human-readable but rather short: three or four words).
  The sixth parameter is the tooltip for this setting (but don't forget to
  copy the same tooltip in the UI file).
  The seventh parameter is the default value for this setting (but don't
  forget to put this default value in the UI which can be relevant in some
  situations).
  The eightth parameter tells Bibble how to handle this setting. You may choose
  several options (all are in the Bibble.h file from the SDK); if you want
  to use several, use a single vertical bar | between them (you may see
  an example in the first line below).
  Most often you can only put OPTION_RESTART which tells that Bibble has to
  restart the computation for this picture when the setting has changed.
  For the "enabled" checkbox I also added a second option to tell Bibble
  it does not need to check dependancies (see the doDependancies function
  above) when the user changes this setting: no other setting has to be
  linked to this action.
*/
__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, BATHSHEBA_ENABLED, "bathsheba_enabled", "Enable Bathsheba", BATHSHEBA_GROUP, "Enable Bathsheba.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
	baf->m_addIntOption( bibbleData->m_plugin, BATHSHEBA_REFWHITE, "bathsheba_refwhite", "Reference illuminant", BATHSHEBA_GROUP, "Set the reference illuminant.\n\nThe standard illuminants are defined as follow:\n\nCIE standard illuminant A is intended to represent typical,\ndomestic, tungsten-filament lighting.\n\nIlluminants B and C are daylight simulators. B served as\na representative of noon sunlight, while C represented\naverage day light.\n\nThe D series of illuminants are constructed to represent\nnatural daylight.\nD50: horizon light;\nD55: mid-morning / mid-afternoon daylight;\nD65: noon daylight;\nD75: North sky daylight.\n\nIlluminant E is an equal-energy radiator.\n\nThe F series of illuminants represent various types of\nfluorescent lighting.\nF2: cool white fluorescent;\nF7: daylight simulator (broadband fluorescent lamp);\nF11: narrow triband fluorescent.", 3, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, BATHSHEBA_SPECTRUM, "bathsheba_spectrum", "Spectrum algorithm", BATHSHEBA_GROUP, "Set the algorithm for computing the spectral power distribution.\nThe XYZ algorithms rest on the CIE matching functions.\nThe D-illuminant computes the spectrum from the temperature.\nThe blackbody uses Planck's law; it is merely experimental here.", 4, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, BATHSHEBA_NORMALIZE, "bathsheba_normalize", "Normalize signal", BATHSHEBA_GROUP, "Normalize the intensity of the reconstructed signal.",1.0, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, BATHSHEBA_YCHANNEL, "bathsheba_ychannel", "Use Y channel", BATHSHEBA_GROUP, "Use Y channel from XYZ for computing\nthe intensity of light (rather than channel\nL from Lab).\nIt has no effect for the XYZ algorithms.", false, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, BATHSHEBA_SELECTCURRENTWL, "bathsheba_selectcurrentwl", "Current wavelength", BATHSHEBA_GROUP, "Select the wavelength.", 21, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, BATHSHEBA_ADJUSTCURRENTWL, "bathsheba_adjustcurrentwl", "Value of current wavelength", BATHSHEBA_GROUP, "Adjust the sensitivity for the currently selected wavelength.", 1.0, OPTION_RESTART);



        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE350, "bathsheba_wl350", "Value for 350nm", BATHSHEBA_GROUP, "Value for 350.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE360, "bathsheba_wl360", "Value for 360nm", BATHSHEBA_GROUP, "Value for 360.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE370, "bathsheba_wl370", "Value for 370nm", BATHSHEBA_GROUP, "Value for 370.nm",0.205479 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE380, "bathsheba_wl380", "Value for 380nm", BATHSHEBA_GROUP, "Value for 380.nm",0.349647 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE390, "bathsheba_wl390", "Value for 390nm", BATHSHEBA_GROUP, "Value for 390.nm",0.448753 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE400, "bathsheba_wl400", "Value for 400nm", BATHSHEBA_GROUP, "Value for 400.nm",0.565196 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE410, "bathsheba_wl410", "Value for 410nm", BATHSHEBA_GROUP, "Value for 410.nm",0.612417 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE420, "bathsheba_wl420", "Value for 420nm", BATHSHEBA_GROUP, "Value for 420.nm",0.636125 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE430, "bathsheba_wl430", "Value for 430nm", BATHSHEBA_GROUP, "Value for 430.nm",0.691809 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE440, "bathsheba_wl440", "Value for 440nm", BATHSHEBA_GROUP, "Value for 440.nm",0.767315 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE450, "bathsheba_wl450", "Value for 450nm", BATHSHEBA_GROUP, "Value for 450.nm",0.83348  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE460, "bathsheba_wl460", "Value for 460nm", BATHSHEBA_GROUP, "Value for 460.nm",0.864277 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE470, "bathsheba_wl470", "Value for 470nm", BATHSHEBA_GROUP, "Value for 470.nm",0.875257 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE480, "bathsheba_wl480", "Value for 480nm", BATHSHEBA_GROUP, "Value for 480.nm",0.874758 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE490, "bathsheba_wl490", "Value for 490nm", BATHSHEBA_GROUP, "Value for 490.nm",0.866943 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE500, "bathsheba_wl500", "Value for 500nm", BATHSHEBA_GROUP, "Value for 500.nm",0.84711  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE510, "bathsheba_wl510", "Value for 510nm", BATHSHEBA_GROUP, "Value for 510.nm",0.837669 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE520, "bathsheba_wl520", "Value for 520nm", BATHSHEBA_GROUP, "Value for 520.nm",0.892555 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE530, "bathsheba_wl530", "Value for 530nm", BATHSHEBA_GROUP, "Value for 530.nm",0.939847 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE540, "bathsheba_wl540", "Value for 540nm", BATHSHEBA_GROUP, "Value for 540.nm",0.946682 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE550, "bathsheba_wl550", "Value for 550nm", BATHSHEBA_GROUP, "Value for 550.nm",0.94984  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE560, "bathsheba_wl560", "Value for 560nm", BATHSHEBA_GROUP, "Value for 560.nm",1        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE570, "bathsheba_wl570", "Value for 570nm", BATHSHEBA_GROUP, "Value for 570.nm",1.05911  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE580, "bathsheba_wl580", "Value for 580nm", BATHSHEBA_GROUP, "Value for 580.nm",1.12005  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE590, "bathsheba_wl590", "Value for 590nm", BATHSHEBA_GROUP, "Value for 590.nm",1.167    ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE600, "bathsheba_wl600", "Value for 600nm", BATHSHEBA_GROUP, "Value for 600.nm",1.19184  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE610, "bathsheba_wl610", "Value for 610nm", BATHSHEBA_GROUP, "Value for 610.nm",1.18863  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE620, "bathsheba_wl620", "Value for 620nm", BATHSHEBA_GROUP, "Value for 620.nm",1.14894  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE630, "bathsheba_wl630", "Value for 630nm", BATHSHEBA_GROUP, "Value for 630.nm",1.15109  ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE640, "bathsheba_wl640", "Value for 640nm", BATHSHEBA_GROUP, "Value for 640.nm",0.952306 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE650, "bathsheba_wl650", "Value for 650nm", BATHSHEBA_GROUP, "Value for 650.nm",0.633293 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE660, "bathsheba_wl660", "Value for 660nm", BATHSHEBA_GROUP, "Value for 660.nm",0.277086 ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE670, "bathsheba_wl670", "Value for 670nm", BATHSHEBA_GROUP, "Value for 670.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE680, "bathsheba_wl680", "Value for 680nm", BATHSHEBA_GROUP, "Value for 680.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE690, "bathsheba_wl690", "Value for 690nm", BATHSHEBA_GROUP, "Value for 690.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE700, "bathsheba_wl700", "Value for 700nm", BATHSHEBA_GROUP, "Value for 700.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE710, "bathsheba_wl710", "Value for 710nm", BATHSHEBA_GROUP, "Value for 710.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE720, "bathsheba_wl720", "Value for 720nm", BATHSHEBA_GROUP, "Value for 720.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE730, "bathsheba_wl730", "Value for 730nm", BATHSHEBA_GROUP, "Value for 730.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE740, "bathsheba_wl740", "Value for 740nm", BATHSHEBA_GROUP, "Value for 740.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE750, "bathsheba_wl750", "Value for 750nm", BATHSHEBA_GROUP, "Value for 750.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE760, "bathsheba_wl760", "Value for 760nm", BATHSHEBA_GROUP, "Value for 760.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE770, "bathsheba_wl770", "Value for 770nm", BATHSHEBA_GROUP, "Value for 770.nm",0        ,OPTION_RESTART);
        baf->m_addDoubleOption(bibbleData->m_plugin, BATHSHEBA_WLVALUE780, "bathsheba_wl780", "Value for 780nm", BATHSHEBA_GROUP, "Value for 780.nm",0        ,OPTION_RESTART);








	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, BATHSHEBA_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
