#include "BathshebaS.h"

BDebugTraceFunction BTestDebug;
#define EPSILON ((double)0.0088564516790356308171716757554635)
#define KAPPA   ((double)903.29629629629629629629629629629630)
#define EPSILONKAPPA ((double)8.0)

void traceCurve(double s1, double s2,
                struct BibblePluginData *bibbleData,
                BSettingsWidgetHandle wh) {
  const char *rgb[175] = {
  "ffab58", "ffaf5d", "ffb262", "ffb669", "ffbb6f", "ffbf76",
  "ffc37d", "ffc684", "ffca8a", "ffce91", "ffd198", "ffd59e",
  "ffd6a0", "ffd9a7", "ffdcae", "ffdfb4", "ffe2ba", "ffe5c0",
  "ffe7c5", "ffeacb", "ffecd0", "ffeed5", "fff0da", "fff2df",
  "fff4e3", "fff6e8", "fff8ec", "fff9f0", "fffbf4", "fffcf8",
  "fffefb", "ffffff", "fbfdff", "f8fbff", "f5f9ff", "f2f7ff",
  "eff5ff", "edf4ff", "eaf2ff", "e8f1ff", "e6efff", "e3eeff",
  "e1ecff", "dfebff", "deeaff", "dce9ff", "dae8ff", "d8e7ff",
  "d7e6ff", "d5e5ff", "d4e4ff", "d2e3ff", "d1e2ff", "d0e1ff",
  "cee0ff", "cddfff", "ccdeff", "cbdeff", "caddff", "c9dcff",
  "c8dcff", "c7dbff", "c6daff", "c5daff", "c4d9ff", "c3d8ff",
  "c2d8ff", "c2d7ff", "c1d7ff", "c0d6ff", "bfd6ff", "bfd5ff",
  "bed5ff", "bdd4ff", "bcd4ff", "bcd3ff", "bbd3ff", "bbd3ff",
  "bad2ff", "b9d2ff", "b9d1ff", "b8d1ff", "b8d1ff", "b7d0ff",
  "b7d0ff", "b6d0ff", "b6cfff", "b5cfff", "b5cfff", "b4ceff",
  "b4ceff", "b3ceff", "b3cdff", "b3cdff", "b2cdff", "b2cdff",
  "b1ccff", "b1ccff", "b1ccff", "b0cbff", "b0cbff", "b0cbff",
  "afcbff", "afcbff", "afcaff", "aecaff", "aecaff", "aecaff",
  "adc9ff", "adc9ff", "adc9ff", "acc9ff", "acc9ff", "acc8ff",
  "acc8ff", "abc8ff", "abc8ff", "abc8ff", "abc7ff", "aac7ff",
  "aac7ff", "aac7ff", "aac7ff", "a9c7ff", "a9c6ff", "a9c6ff",
  "a9c6ff", "a9c6ff", "a8c6ff", "a8c6ff", "a8c6ff", "a8c5ff",
  "a8c5ff", "a7c5ff", "a7c5ff", "a7c5ff", "a7c5ff", "a7c5ff",
  "a6c5ff", "a6c4ff", "a6c4ff", "a6c4ff", "a6c4ff", "a6c4ff",
  "a5c4ff", "a5c4ff", "a5c4ff", "a5c3ff", "a5c3ff", "a5c3ff",
  "a4c3ff", "a4c3ff", "a4c3ff", "a4c3ff", "a4c3ff", "a4c3ff",
  "a4c3ff", "a3c2ff", "a3c2ff", "a3c2ff", "a3c2ff", "a3c2ff",
  "a3c2ff", "a3c2ff", "a3c2ff", "a2c2ff", "a2c2ff", "a2c2ff",
  "a2c1ff", "a2c1ff", "a2c1ff", "a2c1ff", "a2c1ff", "a1c1ff",
  "a1c1ff" };
  s1 *= .5; s2 *= 0.05;
  for (int i=0; i<175;i++) {
    double cct = i*((double)129.3103448275862)+2500.0;
    double cct2 = cct*cct;
    double cct3 = cct2*cct;
    double little_x;
    if(cct<=7000.0) {
      little_x = -4.6070e9/cct3+2.9678e6/cct2+0.09911e3/cct+0.244063;
    } else {
      little_x = -2.0064e9/cct3+1.9018e6/cct2+0.24748e3/cct+0.237040;
    }
    double little_y = -3.0*little_x*little_x+2.870*little_x-0.275;
    double m = ((double)0.0241) + ((double) 0.2562)*little_x
                              - ((double) 0.7341)*little_y;
    double m1 = (((double)-1.3515) - ((double) 1.7703)*little_x
                              + ((double) 5.9114)*little_y)/m;
    double m2 = (((double)0.0300) - ((double) 31.4424)*little_x
                              + ((double) 30.0717)*little_y)/m;
    double value = (s1*m1 + s2*m2 + 1)*64;
    char label[400];
    char label_id[64];
    if(value<0.0) { value = 0.0; }
    else { if(value>128.0) value=128.0; }
    sprintf(label_id,"bathshebaS_curve%d_label",i);
  sprintf(label,"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\"><html><head><meta name=\"qrichtext\" content=\"1\" /></head><body style=\"font-size: 256px\"><table style=\"margin-top: %dpx\" width=\"1px\" bgcolor=\"#%s\"><tbody><tr><td> </td></tr></tbody></table></body></html>",128-((int) (value + .5)),rgb[i]);
    BSettingsWidgetSetLabel(wh, label_id,label);
  }
}

int MainLoopConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg  = BTileImage(pipeData->m_tile);

	// If we don't have an image, we can't do anything
	if (bimg == 0)
	{
		BMSG( "MainLoopConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);
	rw = BImageRowWords(bimg);
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);

	if ( nChan != 3 )
	{
		BMSG("MainLoopConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}
        
        // grab our settings from the widget
        int refwhite;
        double s1,s2,normalize;
        bool ychannel;
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          BATHSHEBA_REFWHITE, bibbleData->m_plugin, refwhite);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          BATHSHEBA_YCHANNEL, bibbleData->m_plugin, ychannel);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          BATHSHEBA_S1, bibbleData->m_plugin, s1);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          BATHSHEBA_S2, bibbleData->m_plugin, s2);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          BATHSHEBA_NORMALIZE, bibbleData->m_plugin, normalize);

        s1 *= .5; s2 *= 0.05;
        // TODO: look for the best value (left)
        normalize *= ((double)650.0);
        if(ychannel) normalize *= (double)260.0;

        double little_x, little_y;
        double rw2 = refwhite*refwhite;
        double rw3 = rw2*refwhite;
        if(refwhite<=7000.0) {
          little_x = -4.6070e9/rw3+2.9678e6/rw2+0.09911e3/refwhite+0.244063;
        } else {
          little_x = -2.0064e9/rw3+1.9018e6/rw2+0.24748e3/refwhite+0.237040;
        }
        little_y = -3.0*little_x*little_x+2.870*little_x-0.275;
        double white_x = little_x/little_y;
        double const white_y = 1.0;
        double white_z = (((double)1.0)-little_x-little_y)/little_y;

	// Convert each scan line including the padding to monochrome
	// May not be necessary to convert the pad, but any processing
        // after us that dips into the pad
	// may cause color artifacts around the edges of the image
	unsigned short *pIn, *pEnd, *pAlpha;
        double R,G,B;

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
#include "core.c"
				*(pIn) = (unsigned short)(R + 0.5f);
				*(pIn + pw) = (unsigned short)32768;
				*(pIn + 2*pw) = (unsigned short)32768;
				pIn += np;
			}
		}
	} else {
	        BImageHandle alpha = pipeData->m_alpha; 
		float ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pAlpha += np;
					continue;
				}
#include "core.c"
				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pIn) = (unsigned short)(R + 0.5f);
					*(pIn + pw) = (unsigned short)32768;
					*(pIn + 2*pw) = (unsigned short)32768;
				} else {
					oval = pIn[0]*ival + R*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[0] = oval;

					oval = pIn[pw]*ival + ((double)32768.0)*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw] = oval;

					oval = pIn[pw+pw]*ival + ((double)32768.0)*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw+pw] = oval;

				}
				pIn += np; pAlpha += np;
			}
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

        traceCurve(0.0,0.0, bibbleData,sw);

	return rc;
}

/*
  The following function is called in order to fix dependancies between
  settings (for instance, if some user has to choose a color with three
  sliders (red, green, blue), you may want to change sliders for some
  Hue, saturation, value model of the same color (HSV) in order to
  allow the user to choose a color by two different (but linked) ways.
  You may see some dependancies with the Simple Color Filter v1.1 (just
  set the color with the sliders, and see the color of the button changing
  in the same time).
  The most known feature handled here is enabling the plugin when the user
  moves a slider.
*/
int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Bathsheba - Do Dependancies");
#endif

	if ((BOptionListContains(od->m_changeList,BATHSHEBA_REFWHITE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_NORMALIZE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_YCHANNEL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_S1,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,BATHSHEBA_S2,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,BATHSHEBA_ENABLED,bibbleData->m_plugin,true);
	}
        double s1,s2;
	BGetOptionDouble(od->m_currentList,BATHSHEBA_S1,bibbleData->m_plugin,s1);
	BGetOptionDouble(od->m_currentList,BATHSHEBA_S2,bibbleData->m_plugin,s2);
	if (BOptionListContains(od->m_changeList,BATHSHEBA_S1,bibbleData->m_plugin)) {
	BGetOptionDouble(od->m_changeList,BATHSHEBA_S1,bibbleData->m_plugin,s1);
        traceCurve(s1,s2,bibbleData,od->m_widget);
        }
	if (BOptionListContains(od->m_changeList,BATHSHEBA_S2,bibbleData->m_plugin)) {
	BGetOptionDouble(od->m_changeList,BATHSHEBA_S2,bibbleData->m_plugin,s2);
        traceCurve(s1,s2,bibbleData,od->m_widget);
        }

        if (BOptionListContains(od->m_changeList, BATHSHEBA_RESETOPT,bibbleData->m_plugin))
        {
		BSetOptionInt(od->m_changeList,BATHSHEBA_REFWHITE,bibbleData->m_plugin,5003);
		BSetOptionDouble(od->m_changeList,BATHSHEBA_S1,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,BATHSHEBA_S2,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,BATHSHEBA_NORMALIZE,bibbleData->m_plugin,1.0);
		BSetOptionBool(od->m_changeList,BATHSHEBA_YCHANNEL,bibbleData->m_plugin,false);
                traceCurve(0.0,0.0, bibbleData,od->m_widget);
        }

	return rc;
}

int hotnessChanged(struct BibblePluginData *bibbleData,struct OptionData *od)
{
	int rc =0;
#ifdef _DEBUG
	BMSG( "Bathsheba Hotness Changed");
#endif
        // Curve
        double s1,s2;
          BGetOptionDouble( od->m_currentList, BATHSHEBA_S1,
            bibbleData->m_plugin, s1);
          BGetOptionDouble( od->m_currentList, BATHSHEBA_S2,
            bibbleData->m_plugin, s2);
        traceCurve(s1,s2, bibbleData,od->m_widget);

	return rc;
}

// This function is a registered button callback
int bathshebaHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Bathsheba - bathshebaHandleButton()" );
#endif

	BMsgBox( "Bathsheba (Solomon version)", "Version 1.0.1\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}
int bathshebaHandleButtonChange( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Bathsheba - bathshebaHandleButtonChange()" );
#endif
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_groupName, BATHSHEBA_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_pluginName, BATHSHEBA_PLUGINNAME);

        // TUTORIAL: choose the stage where you want the plugin
        // to run (you may find the list in the BibblePlugin.h
        // file from the SDK); you may also choose some priority
        // (in order tomake it run before or after another one
        // in the same stage): high priority means that the
        // plugin will run earlier.
        // In order to make some tests, see if your plugin runs
        // before or after the Black and White plugin which has
        // the following settings: PIPE_PLUGIN_CURVES, 200
        // Try a lower and then a higher priority and look at
        // the result.
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, MainLoopConvert, PIPE_PLUGIN_COLORMANAGE2, 225);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, bathshebaHandleButton, BATHSHEBA_BUTTONNAME);
        bibbleData->m_brf.m_registerSettingsButtonCallback(bibbleData->m_plugin, bathshebaHandleButtonChange, BATHSHEBA_RESET,BATHSHEBA_RESETOPT);

        // The function registered below (doDependancies) will be called
        // when a setting will be changed
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies );
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);
	bibbleData->m_brf.m_registerHotnessChanged(bibbleData->m_plugin, hotnessChanged);

	BMSG("Bathsheba plugin loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

/*
  TUTORIAL:
  The following function is very easy to handle but you have to
  spend a little time on it. Add one line for each setting by
  copying or changing existing lines.
	baf->m_addIntOption    (for an integer value) 
	baf->m_addDoubleOption (for a double value)
	baf->m_addBoolOption   (for a boolean value, most often checkboxes)
  The third parameter is the name of the setting in the UI (without the
  suffix: "_slider", "_edit", "_checkbox", etc.
  The fourth parameter is the name of the setting in the "selective settings"
  panel (something human-readable but rather short: three or four words).
  The sixth parameter is the tooltip for this setting (but don't forget to
  copy the same tooltip in the UI file).
  The seventh parameter is the default value for this setting (but don't
  forget to put this default value in the UI which can be relevant in some
  situations).
  The eightth parameter tells Bibble how to handle this setting. You may choose
  several options (all are in the Bibble.h file from the SDK); if you want
  to use several, use a single vertical bar | between them (you may see
  an example in the first line below).
  Most often you can only put OPTION_RESTART which tells that Bibble has to
  restart the computation for this picture when the setting has changed.
  For the "enabled" checkbox I also added a second option to tell Bibble
  it does not need to check dependancies (see the doDependancies function
  above) when the user changes this setting: no other setting has to be
  linked to this action.
*/
__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, BATHSHEBA_ENABLED, "bathshebaS_enabled", "Enable Bathsheba", BATHSHEBA_GROUP, "Enable Bathsheba.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
	baf->m_addIntOption( bibbleData->m_plugin, BATHSHEBA_REFWHITE, "bathshebaS_refwhite", "Reference white", BATHSHEBA_GROUP, "Set the reference white.", 5003, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, BATHSHEBA_S1, "bathshebaS_s1", "Precomputed integral S1", BATHSHEBA_GROUP, "Precomputed integral S1 which will be multiplied by 0.5\n(with S0 being normalized to 1).", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, BATHSHEBA_S2, "bathshebaS_s2", "Precomputed integral S2", BATHSHEBA_GROUP, "Precomputed integral S2 which will be multiplied by 0.05\n(with S0 being normalized to 1).", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, BATHSHEBA_NORMALIZE, "bathshebaS_normalize", "Normalize signal", BATHSHEBA_GROUP, "Normalize the intensity of the reconstructed signal.",1.0, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, BATHSHEBA_YCHANNEL, "bathshebaS_ychannel", "Use Y channel", BATHSHEBA_GROUP, "Use Y channel from XYZ for computing\nthe intensity of light (rather than channel\nL from Lab).", false, OPTION_RESTART);
        baf->m_addBoolOption( bibbleData->m_plugin, BATHSHEBA_RESETOPT, "bathshebaS_resetopt", "Reset settings for Bathsheba", BATHSHEBA_GROUP, "Reset all settings.", false, OPTION_DOESNOTPERSIST|OPTION_NO_DISPLAY);

	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, BATHSHEBA_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
