// TODO: remove initializeOptions
// TODO: prevNeeded
/*

Copyright (c) 2010 Thomas Baruchel

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

#include "LabColors.h"

BDebugTraceFunction BTestDebug;

int LcConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg, dest, alpha;

	if ( pipeData->m_layerNum != 0 )
	{
		alpha = pipeData->m_alpha; 
	} else {
		alpha = 0;
	}
	bimg = dest = BTileImage(pipeData->m_tile);


	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "LcConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("LcConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	// Convert each scan line including the padding to monochrome
	// May not be necessary to convert the pad, but any processing after us that dips into the pad
	// may cause color artifacts around the edges of the image
	unsigned short *pIn, *pOut, *pEnd, *pAlpha;
        double amul,bmul,aadd,badd;
        double achannel,bchannel;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_AMUL, bibbleData->m_plugin, amul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_BMUL, bibbleData->m_plugin, bmul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_AADD, bibbleData->m_plugin, aadd);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_BADD, bibbleData->m_plugin, badd);

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
                          achannel = (((double)*(pIn + pw))-((double)32757.5))
                            *amul
                            +((double)32757.5) +aadd*((double) 8500.0);
                          bchannel = (((double)*(pIn + 2*pw))-((double)32757.5))
                            *bmul
                            +((double)32757.5) +badd*((double) 8500.0);

                          if(achannel<0) achannel = 0;
                          else if (achannel > 65535) achannel = 65535;
                          if(bchannel<0) bchannel = 0;
                          else if (bchannel > 65535) bchannel = 65535;

			  *(pOut + pw) = (unsigned short)(achannel + ((double) 0.5));
			  *(pOut + 2*pw) = (unsigned short)(bchannel + ((double) 0.5));
			  pOut += np; pIn += np;
			}
		}
	} else {
		unsigned int ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pOut += np; pAlpha += np;
					continue;
				}

                          achannel = (((double)*(pIn + pw))-((double)32757.5))
                            *amul
                            +((double)32757.5) +aadd*((double) 8500.0);
                          bchannel = (((double)*(pIn + 2*pw))-((double)32757.5))
                            *bmul
                            +((double)32757.5) +badd*((double) 8500.0);

                          if(achannel<0) achannel = 0;
                          else if (achannel > 65535) achannel = 65535;
                          if(bchannel<0) bchannel = 0;
                          else if (bchannel > 65535) bchannel = 65535;

				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pOut + pw) = (unsigned short)(achannel + ((double) 0.5));
					*(pOut + 2*pw) = (unsigned short)(bchannel + ((double) 0.5));
				}
				else
				{
					//	We should be blending with 'pOut' (dest) - or in other words
					//	the tile's (pipeData->m_tile) image.  This will contain the 
					//	results of the previous layers and must be used in order
					//	for the blending to be correct.
					//	We don't use pIn (bimg) here because for this plugin that is
					//	a copy of the previous filter's end result and will not contain
					//	the results of previous layer processing for this filter.

					oval = pOut[pw] * ival + achannel * aval + ((double) 0.5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw] = oval;

					oval = pOut[pw+pw] * ival + bchannel * aval + ((double) 0.5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw+pw] = oval;

				}
				pOut += np; pIn += np; pAlpha += np;
			}
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	return rc;
}

int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

	// Get a handle to our widget... if we can't find it, we can't initialize our options
        /*
	BSettingsWidgetHandle wh = od->m_widget;
	if ( NULL == wh ) return -1;
        */

#ifdef _DEBUG
	BMSG("Channel Mixer Plugin - Do Dependencies");
#endif

	if ((BOptionListContains(od->m_changeList,LC_AMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,LC_BMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,LC_AADD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,LC_BADD,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,LC_ENABLED,bibbleData->m_plugin,true);
	}

	return rc;
}

// This function is a registered button callback
int LcHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Lab Colors Plugin - LcHandleButton()" );
#endif

	BMsgBox( "Lab Colors", "Version 1.0\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_groupName, LC_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_pluginName, LC_PLUGINNAME);

        // 300 after Inda, before BW
        // 350 before Inda, before BW, after PresetToni
        // 500 after Preset Toni
        // 650 before SplitToni, after Preset Toni
        // 700 before Preset Toni and SplitToni after Harry
        // 1000 after Harry
        // 5000 after Harry
        // 50000 after Harry
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, LcConvert, PIPE_PLUGIN_COLORMANAGE2, 200);
	//bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, LcHandleButton, LC_BUTTONNAME);

	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies);

	BMSG("LabColors plug-in loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, LC_ENABLED, "labcolorsenabled", "Enable Lab Colors", LC_GROUP, "Enable Lab Colors.", false, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LC_AMUL, "achannelmul", "Factor for channel a", LC_GROUP, "Multiplicative factor for channel a.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LC_BMUL, "bchannelmul", "Factor for channel b", LC_GROUP, "Multiplicative factor for channel b.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LC_AADD, "achanneladd", "Filter Green / Red", LC_GROUP, "Apply some green/red color filter.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LC_BADD, "bchanneladd", "Filter Blue / Yellow", LC_GROUP, "Apply some blue/yellow color filter.", 0.0, OPTION_RESTART);
	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, LC_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
