/*

Copyright (c) 2010 Thomas Baruchel

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

#include "LabContrast.h"

BDebugTraceFunction BTestDebug;

int LcConvert0(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
        int rc = 0;

        bool rgb;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, LCO_RGB, bibbleData->m_plugin, rgb);
        if(!rgb) {

	// Grab the image data and info
	BImageHandle bimg, dest, alpha;

	if ( pipeData->m_layerNum != 0 )
	{
		alpha = pipeData->m_alpha; 
	} else {
		alpha = 0;
	}
	bimg = dest = BTileImage(pipeData->m_tile);

        double exposure,gamma,middlegray,contrast;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_EXPOSURE, bibbleData->m_plugin, exposure);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_DETAILS, bibbleData->m_plugin, gamma);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_MIDDLEGRAY, bibbleData->m_plugin, middlegray);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_CONTRAST, bibbleData->m_plugin, contrast);
        if(gamma<0) {
          gamma = ((double) 1.0) - gamma;
        } else {
          gamma = ((double)1.0)/(((double)1.0)+gamma);
        }
        middlegray = ((double) 1.0) - middlegray;

	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "LcConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("LcConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	unsigned short *pIn, *pOut, *pEnd, *pAlpha;
        double lchannel;

        double beta = contrast*middlegray;
        double c1 = ((double) 1.0)/(((double) 1.0)+exp(beta));
        double c2 = ((double) 1.0)/(((double) 1.0)+exp(beta-contrast))-c1;
	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
                            if(contrast>0.005) {
                            lchannel = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            } else {
                            lchannel = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma);
                            }

                            if(lchannel<0) lchannel = 0.0;
                            else if (lchannel > 65535.0) lchannel = 65535.0;

			    *(pOut) = (unsigned short)(lchannel + ((double) 0.5));
			    pOut += np; pIn += np;
			}
		}
	} else {
		unsigned int ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pOut += np; pAlpha += np;
					continue;
				}

                            if(contrast>0.005) {
                            lchannel = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            } else {
                            lchannel = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma);
                            }

                            if(lchannel<0) lchannel = 0.0;
                            else if (lchannel > 65535.0) lchannel = 65535.0;

				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pOut) = (unsigned short)(lchannel+ 0.5f);
				} else {
			    	oval = pOut[0] * ival + lchannel * aval + ((double) 0.5);
			    	oval >>= 16;
			    	oval = oval & 0x0000ffff;

			    	if (oval < 0)
			    		oval = 0;
			    	if (oval > 65535)
			    		oval = 65535;
			    	pOut[0] = oval;
			    }

			pOut += np; pIn += np; pAlpha += np;
			}
		}
	}
        }
	return rc;
}

int LcConvert1(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

        bool rgb;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, LCO_RGB, bibbleData->m_plugin, rgb);
        if(rgb) {

	// Grab the image data and info
	BImageHandle bimg, dest, alpha;

	if ( pipeData->m_layerNum != 0 )
	{
		alpha = pipeData->m_alpha; 
	} else {
		alpha = 0;
	}
	bimg = dest = BTileImage(pipeData->m_tile);

        double exposure,gamma,middlegray,contrast;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_EXPOSURE, bibbleData->m_plugin, exposure);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_DETAILS, bibbleData->m_plugin, gamma);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_MIDDLEGRAY, bibbleData->m_plugin, middlegray);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LCO_CONTRAST, bibbleData->m_plugin, contrast);
        if(gamma<0) {
          gamma = ((double) 1.0) - gamma;
        } else {
          gamma = ((double)1.0)/(((double)1.0)+gamma);
        }
        middlegray = ((double) 1.0) - middlegray;

	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "LcConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("LcConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	unsigned short *pIn, *pOut, *pEnd, *pAlpha;
        double R,G,B;

        double beta = contrast*middlegray;
        double c1 = ((double) 1.0)/(((double) 1.0)+exp(beta));
        double c2 = ((double) 1.0)/(((double) 1.0)+exp(beta-contrast))-c1;
	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
                            if(contrast>0.005) {
                            R = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            G = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn+pw))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            B = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn+2*pw))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            } else {
                            R = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma);
                            G = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn+pw))/((double)65535.0)*exposure,
                              gamma);
                            B = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn+2*pw))/((double)65535.0)*exposure,
                              gamma);
                            }

                            if(R<0) R = 0.0;
                            else if (R > 65535.0) R = 65535.0;
                            if(G<0) G = 0.0;
                            else if (G > 65535.0) G = 65535.0;
                            if(B<0) B = 0.0;
                            else if (B > 65535.0) B = 65535.0;
				*(pOut) = (unsigned short)(R + 0.5f);
				*(pOut + pw) = (unsigned short)(G + 0.5f);
				*(pOut + 2*pw) = (unsigned short)(B + 0.5f);
				pIn += np; pOut += np;
			}
		}
	} else {
		unsigned int ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pOut += np; pAlpha += np;
					continue;
				}

                            if(contrast>0.005) {
                            R = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            G = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn+pw))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            B = ((double) 65535.0)*
                            (((double) 1.0) / (
                            ((double) 1.0)+
                            exp(beta-contrast*pow(
                              ((double)*(pIn+2*pw))/((double)65535.0)*exposure,
                              gamma))) - c1)/c2;
                            } else {
                            R = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn))/((double)65535.0)*exposure,
                              gamma);
                            G = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn+pw))/((double)65535.0)*exposure,
                              gamma);
                            B = ((double) 65535.0)*
                            pow(
                              ((double)*(pIn+2*pw))/((double)65535.0)*exposure,
                              gamma);
                            }

                            if(R<0) R = 0.0;
                            else if (R > 65535.0) R = 65535.0;
                            if(G<0) G = 0.0;
                            else if (G > 65535.0) G = 65535.0;
                            if(B<0) B = 0.0;
                            else if (B > 65535.0) B = 65535.0;

				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pOut) = (unsigned short)(R + 0.5f);
					*(pOut + pw) = (unsigned short)(G + 0.5f);
					*(pOut + 2*pw) = (unsigned short)(B + 0.5f);
				} else {
					oval = pOut[0]*ival + R*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[0] = oval;

					oval = pOut[pw]*ival + G*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw] = oval;

					oval = pOut[pw+pw]*ival + B*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw+pw] = oval;

				}

			pOut += np; pIn += np; pAlpha += np;
			}
		}
	}
        }
	return rc;
}


// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	return rc;
}

int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

	// Get a handle to our widget... if we can't find it, we can't initialize our options
        /*
	BSettingsWidgetHandle wh = od->m_widget;
	if ( NULL == wh ) return -1;
        */

#ifdef _DEBUG
	BMSG("Lab Contrast - Do Dependancies");
#endif
	if ((BOptionListContains(od->m_changeList,LCO_EXPOSURE,bibbleData->m_plugin))||(BOptionListContains(od->m_changeList,LCO_DETAILS,bibbleData->m_plugin))||(BOptionListContains(od->m_changeList,LCO_MIDDLEGRAY,bibbleData->m_plugin))||(BOptionListContains(od->m_changeList,LCO_CONTRAST,bibbleData->m_plugin))||(BOptionListContains(od->m_changeList,LCO_RGB,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,LCO_ENABLED,bibbleData->m_plugin,true);
	}

	return rc;
}

// This function is a registered button callback
int LcHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Lab Contrast - LcHandleButton()" );
#endif

	BMsgBox( "Lab Contrast", "Version 1.0\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_groupName, LCO_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_pluginName, LCO_PLUGINNAME);

	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, LcConvert0, PIPE_PLUGIN_COLORMANAGE2, 200);
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, LcConvert1, PIPE_PLUGIN_COLOR, 800);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, LcHandleButton, LCO_BUTTONNAME);

	// Settings about to change
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies);

	BMSG("LabContrast plug-in loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, LCO_ENABLED, "labcontrastenabled", "Enable Lab Contrast", LCO_GROUP, "Enable Lab Contrast.\n\nYou should first disable the AutoLevel option.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
	baf->m_addDoubleOption( bibbleData->m_plugin, LCO_EXPOSURE, "lco_exposure", "Exposure", LCO_GROUP, "Adjust the general exposure of the picture.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LCO_DETAILS, "lco_details", "Details", LCO_GROUP, "Focus on some details when adjusting light.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LCO_MIDDLEGRAY, "lco_middlegray", "Middle Gray", LCO_GROUP, "Set the center of the contrast curve.", 0.5, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LCO_CONTRAST, "lco_contrast", "Contrast", LCO_GROUP, "Set the strength of the contrast curve.", 0.0, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, LCO_RGB, "labcontrastrgb", "Run on gamma-corrected RGB", LCO_GROUP, "The plugin is first intended to work in Lab colorspace.\nIf you find this behaviour too strange but still want to\nuse the sigmoidal contrast feature of the plugin, you\nmay want to enable this option.", true, OPTION_RESTART);
	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, LCO_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
