/*

This plugin has been written by Thomas Baruchel
from the code of the Black and White plugin in Bibble SDK.

Most of the comments make no sense here (they were comments
from the Black and White plugin).

*/

#include "LabCurves.h"

BDebugTraceFunction BTestDebug;

int LcConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg, dest;

	dest = BTileImage(pipeData->m_tile);
	if ( pipeData->m_layerNum != 0 )
		bimg = pipeData->m_prev; // Output from previous step
	else
		bimg = dest;

        double amplitude;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_AMPLITUDE, bibbleData->m_plugin, amplitude);

	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "LcConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("LcConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	unsigned short *pIn, *pOut, *pEnd, *pAlpha;

	for( int y = -padh; y < h+padh; ++y ) {
		pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
		pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
		pEnd = pIn + (w+padw*2)*np;
		while( pIn < pEnd ) {
                  double rgbr = ((double) *(pIn))/((double) 65535.0);
                  double rgbg = ((double) *(pIn + pw))/((double) 65535.0);
                  double rgbb = ((double) *(pIn + 2*pw))/((double) 65535.0);
		  pIn += np;

                  // ProPhotoRGB is assumed
                  const double gamma = 1.8;
                  rgbr = pow(rgbr,gamma);
                  rgbg = pow(rgbg,gamma);
                  rgbb = pow(rgbb,gamma);
                  double X = .7976749*rgbr
                             + 0.1351917*rgbg
                             + .0313534*rgbb;
                  double Y = .2880402*rgbr
                             + .7118741*rgbg
                             + .0000857*rgbb;
                  double Z = .82521*rgbb;

                  // values from http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_Lab.html
                  const double epsilon = ((double) 216.0)/((double) 24389.0);
                  const double kappa   = ((double) 24389.0)/((double) 27.0);

                  // D50 white point tristimulus values
                  // values from http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
                  const double Xn = .96422;
                  const double Yn = 1.0;
                  const double Zn = 0.82521;

                  double xr = X/Xn;
                  double yr = Y/Yn;
                  double zr = Z/Zn;

                  double fx,fy,fz;
                  if(xr > epsilon)
                    fx = pow(xr,0.33333333333333333333333333);
                  else
                    fx = (kappa*xr+((double) 16.0))/((double) 116.0);

                  if(yr > epsilon)
                    fy = pow(yr,0.33333333333333333333333333);
                  else
                    fy = (kappa*yr+((double) 16.0))/((double) 116.0);

                  if(zr > epsilon)
                    fz = pow(zr,0.33333333333333333333333333);
                  else
                    fz = (kappa*zr+((double) 16.0))/((double) 116.0);

                  double labl,laba,labb;
                  labl = ((double) 116.0)*fy-((double) 16.0);
                  laba = ((double) 500.0)*(fx-fy);
                  labb = ((double) 200.0)*(fy-fz);

                  labl *= ((double) 65535.0)/((double) 100.0);
                  laba = ((double) 65535.0)/(((double) 2.0)*amplitude)*(laba+amplitude);
                  labb = ((double) 65535.0)/(((double) 2.0)*amplitude)*(labb+amplitude);

                  if(labl<((double) 0.0)) labl = ((double) 0.0);
                  else if (labl>((double) 65535.0)) labl = ((double) 65535.0);
                  if(laba<((double) 0.0)) laba = ((double) 0.0);
                  else if (laba>((double) 65535.0)) laba = ((double) 65535.0);
                  if(labb<((double) 0.0)) labb = ((double) 0.0);
                  else if (labb>((double) 65535.0)) labb = ((double) 65535.0);

		  *(pOut) = (unsigned short)(labl+0.5);
		  *(pOut + pw) = (unsigned short)(laba+0.5);
		  *(pOut + 2*pw) = (unsigned short)(labb+0.5);
		  pOut += np;
		}
	}
	return rc;
}

int LcConvertBack(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg, dest;

	dest = BTileImage(pipeData->m_tile);
	if ( pipeData->m_layerNum != 0 )
		bimg = pipeData->m_prev; // Output from previous step
	else
		bimg = dest;

        double amplitude;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          LC_AMPLITUDE, bibbleData->m_plugin, amplitude);


	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "LcConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("LcConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	unsigned short *pIn, *pOut, *pEnd, *pAlpha;

	for( int y = -padh; y < h+padh; ++y ) {
		pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
		pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
		pEnd = pIn + (w+padw*2)*np;
		while( pIn < pEnd ) {
                  double labl = ((double) *(pIn))/((double) 65535.0)*((double) 100.0);
                  double laba = ((double) *(pIn + pw))/((double) 65535.0)*((double) 2.0)*amplitude - amplitude;
                  double labb = ((double) *(pIn + 2*pw))/((double) 65535.0)*((double) 2.0)*amplitude - amplitude;
		  pIn += np;

                  double fy = (labl+((double) 16.0))/((double) 116.0);
                  double fx = laba/((double) 500.0)+fy;
                  double fz = fy - labb/((double) 200.0);

                  // values from http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_Lab.html
                  const double epsilon = ((double) 216.0)/((double) 24389.0);
                  const double kappa   = ((double) 24389.0)/((double) 27.0);

                  double fx3 = pow(fx,((double) 3.0));
                  double fz3 = pow(fz,((double) 3.0));
                  double xr,yr,zr;
                  if(fx3 > epsilon)
                    xr = fx3;
                  else
                    xr = (((double) 116.0)*fx-((double) 16.0))/kappa;
                  if(labl>(kappa*epsilon))
                    yr = pow(
                           (labl+((double) 16.0))/((double) 116.0),
                           ((double) 3.0));
                  else
                    yr = labl/kappa;
                  if(fz3 > epsilon)
                    zr = fz3;
                  else
                    zr = (((double) 116.0)*fz-((double) 16.0))/kappa;

                  // D50 white point tristimulus values
                  // values from http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
                  const double Xn = .96422;
                  const double Yn = 1.0;
                  const double Zn = 0.82521;

                  double X = xr*Xn;
                  double Y = yr*Yn;
                  double Z = zr*Zn;

                  double rgbr = 1.3459433*X-.2556075*Y-.0511118*Z;
                  double rgbg = -.5445989*X+1.5081673*Y+.0205351*Z;
                  double rgbb = 1.2118128*Z;

                  const double igamma = ((double) 1.0)/((double) 1.8);
                  rgbr = pow(rgbr,igamma) * ((double) 65535.0);
                  rgbg = pow(rgbg,igamma) * ((double) 65535.0);
                  rgbb = pow(rgbb,igamma) * ((double) 65535.0);

                  if(rgbr<((double) 0.0)) rgbr = ((double) 0.0);
                  else if (rgbr>((double) 65535.0)) rgbr = ((double) 65535.0);
                  if(rgbg<((double) 0.0)) rgbg = ((double) 0.0);
                  else if (rgbg>((double) 65535.0)) rgbg = ((double) 65535.0);
                  if(rgbb<((double) 0.0)) rgbb = ((double) 0.0);
                  else if (rgbb>((double) 65535.0)) rgbb = ((double) 65535.0);

		  *(pOut) = (unsigned short)(rgbr+0.5);
		  *(pOut + pw) = (unsigned short)(rgbg+0.5);
		  *(pOut + 2*pw) = (unsigned short)(rgbb+0.5);
		  pOut += np;
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	return rc;
}
int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

	// Get a handle to our widget... if we can't find it, we can't initialize our options
        /*
	BSettingsWidgetHandle wh = od->m_widget;
	if ( NULL == wh ) return -1;
        */

#ifdef _DEBUG
	BMSG("Lab Curves - Do Dependancies");
#endif
	if (BOptionListContains(od->m_changeList,LC_AMPLITUDE,bibbleData->m_plugin))
	{
		BSetOptionBool(od->m_changeList,LC_ENABLED,bibbleData->m_plugin,true);
	}

	return rc;
}

int	settingsChanged(struct BibblePluginData *bibbleData,struct OptionData *od)
{
	int rc = 0;
#ifdef _DEBUG
	BMSG("Lab Curves - Settings Changed");
#endif
	return rc;
}

int	settingsAboutToChange(struct BibblePluginData *bibbleData,struct OptionData *od)
{	
	// This function is called when any option change has been broadcasted
	// If any of the plug-in settings interact with or depend on any other
	// settings, change them here. If an option is changed within this function,
	// be sure to add it to the option change list, or that change will not  
	// propagate into the option system.
	int rc = 0;
#ifdef _DEBUG
	BMSG("Lab Curves - Settings About To Change");
#endif

	// We need to be able to read in our options
	// (Not really, we can still populate the combo box if there is no image open!)
	//if ( NULL == od->m_options ) return -1;
	
	// Get a handle to our widget... if we can't find it, we can't initialize our options
        /*
	BSettingsWidgetHandle wh = od->m_widget;
	if ( NULL == wh ) return -1;
        */

	return rc;
}

int hotnessChanged(struct BibblePluginData *bibbleData,struct OptionData *od)
{
	int rc = 0;
#ifdef _DEBUG
	BMSG( "Lab Curves - Hotness Changed");
#endif
	return rc;
}

// This function is a registered button callback
int LcHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Lab Curves - LcHandleButton()" );
#endif

	BMsgBox( "Lab Curves", "Written by Thomas Baruchel");
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_groupName, LC_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_pluginName, LC_PLUGINNAME);

	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, LcConvert, PIPE_PLUGIN_CURVES, 1);
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, LcConvertBack, PIPE_PLUGIN_FINAL, 9999);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, LcHandleButton, LC_BUTTONNAME);

	// Settings about to change
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies);

	BMSG("LabCurves plug-in loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, LC_ENABLED, "lcenabled", "Enable Lab Curves", LC_GROUP, "Enable Lab Curves.\n\nThe working space has to be set to ProPhoto RGB.", false, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, LC_AMPLITUDE, "lcamplitude", "Amplitude for a and b channels", LC_GROUP, "Since many pictures will probably not use the\nwhole ProPhoto RGB colorspace gamut, you may\nfind more convenient to reduce the amplitude\nused bu the two a and b channels.", 186.7, OPTION_RESTART);
	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = true;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, LC_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
