/*

Copyright (c) 2010 Thomas Baruchel

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

#include "SelectColor.h"

BDebugTraceFunction BTestDebug;
#define EPSILON ((double)0.0088564516790356308171716757554635)
#define KAPPA   ((double)903.29629629629629629629629629629630)
#define EPSILONKAPPA ((double)8.0)
#define PI ((double) 3.1415926535897932384626433832795028841971693993751)
#define DPI ((double)6.2831853071795864769252867665590057683943387987502)
#define HPI ((double)1.5707963267948966192313216916397514420985846996876)
#define DEGTORAD ((double)0.0174532925199432957692369076848861271344287188854)

int ScConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg = BTileImage(pipeData->m_tile);

	// If we don't have an image, we can't do anything
	if (bimg == 0)
	{
		BMSG( "ScConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

#include "tables.h"

        int dw,hue;
        double calibrate,radius,xmul,ymul,amul,bmul,aadd,badd,xadd,yadd,
               exposure,gamma,middlegray,contrast;
        bool preview,fullrange,lchannel;
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          SC_DW, bibbleData->m_plugin, dw);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_CALIBRATE, bibbleData->m_plugin, calibrate);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_RADIUS, bibbleData->m_plugin, radius);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          SC_FULLRANGE, bibbleData->m_plugin, fullrange);
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          SC_PREVIEW, bibbleData->m_plugin, preview);
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          SC_HUE, bibbleData->m_plugin, hue);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_XMUL, bibbleData->m_plugin, xmul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_YMUL, bibbleData->m_plugin, ymul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_AMUL, bibbleData->m_plugin, amul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_BMUL, bibbleData->m_plugin, bmul);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_XADD, bibbleData->m_plugin, xadd); xadd *= 0.25;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_YADD, bibbleData->m_plugin, yadd); yadd *= 0.25;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_AADD, bibbleData->m_plugin, aadd); aadd *= 5.0;
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_BADD, bibbleData->m_plugin, badd); badd *= 5.0;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          SC_LCHANNEL, bibbleData->m_plugin, lchannel);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_EXPOSURE, bibbleData->m_plugin, exposure);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_GAMMA, bibbleData->m_plugin, gamma);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_MIDDLEGRAY, bibbleData->m_plugin, middlegray);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SC_CONTRAST, bibbleData->m_plugin, contrast);
        double proximity_setting = 0.25+radius*0.35;
        double proximity_gamma;
        // 0 - 0.5   ->    4 - 1
        // 0.5 - 1   ->    1 - 1/4
        if(radius<0.5) { proximity_gamma = 4.0-6.0*radius; }
        else { proximity_gamma = 1.0/(6.0*radius-2.0); }

        if(gamma<0) {
          gamma = ((double) 1.0) - gamma;
        } else {
          gamma = ((double)1.0)/(((double)1.0)+gamma);
        }
        middlegray = ((double) 1.0) - middlegray;
        double beta = contrast*middlegray;
        double c1 = ((double) 1.0)/(((double) 1.0)+exp(beta));
        double c2 = ((double) 1.0)/(((double) 1.0)+exp(beta-contrast))-c1;

        double cct;
        if(calibrate<0) { cct = 5003+calibrate*1003; }
        else { cct = 5003+calibrate*19997; }
        double little_x_white,little_y_white;
        double cct2 = cct*cct;
        double cct3 = cct2*cct;
        if(cct<=7000.0) {
          little_x_white = -4.6070e9/cct3+2.9678e6/cct2+0.09911e3/cct+0.244063;
        } else {
          little_x_white = -2.0064e9/cct3+1.9018e6/cct2+0.24748e3/cct+0.237040;
        }
        little_y_white = -3.0*little_x_white*little_x_white+2.870*little_x_white-0.275;
        double white_x = little_x_white/little_y_white;
        const double white_y = 1.0;
        double white_z = (1.0-little_x_white-little_y_white)/little_y_white;

        // precompute table of arctangents
        // (according to reference white)
        dw -= 360;
        /*
        double dw_little_x = cie_x[dw]/(cie_x[dw]+cie_y[dw]+cie_z[dw]);
        double dw_little_y = cie_y[dw]/(cie_x[dw]+cie_y[dw]+cie_z[dw]);
        double dw_arct = atan2(dw_little_y-little_y_white,
                               dw_little_x-little_x_white);
        */
        double dw_arct = atan2(
          cie_y[dw]/(cie_x[dw]+cie_y[dw]+cie_z[dw]) - little_y_white,
          cie_x[dw]/(cie_x[dw]+cie_y[dw]+cie_z[dw]) - little_x_white );


	int w,h,nChan,pw,rw,padw,padh,np;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);	// plane words - how far to get to the next plane
	rw = BImageRowWords(bimg);		// number of unsigned shorts in a row (includes padding)
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);		// increment to get to the next pixel

	if ( nChan != 3 )
	{
		BMSG("ScConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}

	// Convert each scan line including the padding to monochrome
	// May not be necessary to convert the pad, but any processing after us that dips into the pad
	// may cause color artifacts around the edges of the image
	unsigned short *pIn, *pEnd, *pAlpha;

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
			while( pIn < pEnd ) {
#include "core.c"
                        *(pIn) = (unsigned short) L;
                        *(pIn+pw) = (unsigned short) a;
                        *(pIn+pw+pw) = (unsigned short) b;

                          pIn += np;
			}
		}
	} else {
	        BImageHandle alpha = pipeData->m_alpha; 
		unsigned int ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{

			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np;
                                        pAlpha += np;
					continue;
				}

#include "core.c"

				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pIn) = (unsigned short)(L + ((double) 0.5));
					*(pIn + pw) = (unsigned short)(a + ((double) 0.5));
					*(pIn + 2*pw) = (unsigned short)(b + ((double) 0.5));
				}
				else
				{
					oval = pIn[0] * ival + L * aval + ((double) 0.5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[0] = oval;

					oval = pIn[pw] * ival + a * aval + ((double) 0.5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw] = oval;

					oval = pIn[pw+pw] * ival + b * aval + ((double) 0.5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pIn[pw+pw] = oval;

				}
                                pIn += np; pAlpha += np;
			}
		}
	}
	return rc;
}

int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Select Color - Do Dependencies");
#endif

	if ((BOptionListContains(od->m_changeList,SC_DW,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_PREVIEW,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_RADIUS,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_CALIBRATE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_FULLRANGE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_HUE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_XMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_YMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_AMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_BMUL,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_XADD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_YADD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_AADD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_BADD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_MIDDLEGRAY,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_CONTRAST,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_EXPOSURE,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_GAMMA,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SC_LCHANNEL,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,SC_ENABLED,bibbleData->m_plugin,true);
	}
        if (BOptionListContains(od->m_changeList, SC_RESETOPT,bibbleData->m_plugin))
        {
		BSetOptionInt(od->m_changeList,SC_DW,bibbleData->m_plugin,560);
		BSetOptionBool(od->m_changeList,SC_PREVIEW,bibbleData->m_plugin,false);
		BSetOptionBool(od->m_changeList,SC_FULLRANGE,bibbleData->m_plugin,false);
		BSetOptionDouble(od->m_changeList,SC_RADIUS,bibbleData->m_plugin,0.5);
		BSetOptionDouble(od->m_changeList,SC_CALIBRATE,bibbleData->m_plugin,0.0);
		BSetOptionInt(od->m_changeList,SC_HUE,bibbleData->m_plugin,0);
		BSetOptionDouble(od->m_changeList,SC_XMUL,bibbleData->m_plugin,1.0);
		BSetOptionDouble(od->m_changeList,SC_YMUL,bibbleData->m_plugin,1.0);
		BSetOptionDouble(od->m_changeList,SC_AMUL,bibbleData->m_plugin,1.0);
		BSetOptionDouble(od->m_changeList,SC_BMUL,bibbleData->m_plugin,1.0);
		BSetOptionDouble(od->m_changeList,SC_XADD,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,SC_YADD,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,SC_AADD,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,SC_BADD,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,SC_MIDDLEGRAY,bibbleData->m_plugin,0.5);
		BSetOptionDouble(od->m_changeList,SC_CONTRAST,bibbleData->m_plugin,0.0);
		BSetOptionDouble(od->m_changeList,SC_EXPOSURE,bibbleData->m_plugin,1.0);
		BSetOptionDouble(od->m_changeList,SC_GAMMA,bibbleData->m_plugin,0.0);
		BSetOptionBool(od->m_changeList,SC_LCHANNEL,bibbleData->m_plugin,false);
        }

	return rc;
}

// This function is a registered button callback
int ScHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Select Color - ScHandleButton()" );
#endif

	BMsgBox( "Select Color", "Version 0.99 beta\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}
int ScHandleButtonChange( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Select Color - ScHandleButtonChange()" );
#endif
	return rc;
}
__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_groupName, SC_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii(bibbleData->m_pluginName, SC_PLUGINNAME);

        // 300 after Inda, before BW
        // 350 before Inda, before BW, after PresetToni
        // 500 after Preset Toni
        // 650 before SplitToni, after Preset Toni
        // 700 before Preset Toni and SplitToni after Harry
        // 1000 after Harry
        // 5000 after Harry
        // 50000 after Harry
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, ScConvert, PIPE_PLUGIN_COLORMANAGE2, 240);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, ScHandleButton, SC_BUTTONNAME);
        bibbleData->m_brf.m_registerSettingsButtonCallback(bibbleData->m_plugin, ScHandleButtonChange, SC_RESET,SC_RESETOPT);

	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies);

	BMSG("SelectColor plug-in loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption(bibbleData->m_plugin, SC_ENABLED, "selectcolor_enabled", "Enable Select Color", SC_GROUP, "Enable Select Color.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
        baf->m_addBoolOption(bibbleData->m_plugin, SC_RESETOPT, "labcolors_resetopt", "Reset settings for LabColors", SC_GROUP, "Reset all settings.", false, OPTION_DOESNOTPERSIST|OPTION_NO_DISPLAY);
	baf->m_addIntOption(bibbleData->m_plugin, SC_DW, "selectcolor_dw", "Color", SC_GROUP, "Select a color by its dominant wavelength.", 560, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_CALIBRATE, "selectcolor_calibrate", "Calibrate", SC_GROUP, "Calibrate the colors (according to the kind of light).", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_RADIUS, "selectcolor_radius", "Radius", SC_GROUP, "Set the radius for selecting colors.", 0.5, OPTION_RESTART);
	baf->m_addBoolOption( bibbleData->m_plugin, SC_FULLRANGE, "selectcolor_fullrange", "Full range", SC_GROUP, "Work on all pixels (with more effect on selected color).", false, OPTION_RESTART);
	baf->m_addBoolOption(bibbleData->m_plugin, SC_PREVIEW, "selectcolor_preview", "Preview mode", SC_GROUP, "Show the selected pixels.", false, OPTION_RESTART);
	baf->m_addIntOption(bibbleData->m_plugin, SC_HUE, "selectcolor_hue", "Hue", SC_GROUP, "Adjust the hue.", 0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_XMUL, "selectcolor_xmul", "Channel x", SC_GROUP, "Change saturation of channel x.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_YMUL, "selectcolor_ymul", "Channel y", SC_GROUP, "Change saturation of channel y.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_AMUL, "selectcolor_achannelmul", "Channel a", SC_GROUP, "Change saturation of channel a.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_BMUL, "selectcolor_bchannelmul", "Channel b", SC_GROUP, "Change saturation of channel b.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_AADD, "selectcolor_achanneladd", "Filter cyan / orange", SC_GROUP, "Apply a cyan / orange filter on channel a.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_BADD, "selectcolor_bchanneladd", "Filter magenta / green", SC_GROUP, "Apply a magenta / green filter on channel b.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_XADD, "selectcolor_xchanneladd", "Filter green / red", SC_GROUP, "Apply a green / red filter on channel x.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_YADD, "selectcolor_ychanneladd", "Filter blue / yellow", SC_GROUP, "Apply a blue / yellow filter on channel y.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_MIDDLEGRAY, "selectcolor_middlegray", "Middlegray", SC_GROUP, "Define the center of the sigmoidal curve.", 0.5, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_CONTRAST, "selectcolor_contrast", "Contrast", SC_GROUP, "Define the strength of the contrast.", 0.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_EXPOSURE, "selectcolor_exposure", "Exposure", SC_GROUP, "Adjust the exposure.", 1.0, OPTION_RESTART);
	baf->m_addDoubleOption(bibbleData->m_plugin, SC_GAMMA, "selectcolor_gamma", "Gamma", SC_GROUP, "Apply a gamma curve.", 0.0, OPTION_RESTART);
	baf->m_addBoolOption(bibbleData->m_plugin, SC_LCHANNEL, "selectcolor_lchannel", "Use the L channel", SC_GROUP, "Apply the settings on the L channel from Cie Lab\n(rather than on the Y channel from Cie xyY).", false, OPTION_RESTART);
	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, SC_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
