// TUTORIAL: don't forget to change this:
#include "SimpleFrame.h"

BDebugTraceFunction BTestDebug;


int MainLoopConvert(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData) {
	int rc = 0;

	// Grab the image data and info
	BImageHandle bimg, dest, alpha;

	if ( pipeData->m_layerNum != 0 )
	{
		alpha = pipeData->m_alpha; 
	} else {
		alpha = 0;
	}
	bimg = dest = BTileImage(pipeData->m_tile);

	// If we don't have an image, we can't do anything
	if ( bimg == 0 || dest == 0 )
	{
		BMSG( "MainLoopConvert() failed - NULL image");
		return -1;
	}

	pipeData->m_blend = false; 

	int w,h,nChan,pw,rw,padw,padh,np;
        int y_offset,x_offset,y_maxsize,x_maxsize;
        unsigned int tile_scale;

	w = BImageWidth(bimg);
	h = BImageHeight(bimg);
	nChan = BImageNumChannels(bimg);
	pw = BImagePlaneWords(bimg);
	rw = BImageRowWords(bimg);
	padw = bibbleData->m_bif.m_padWidth( bimg );
	padh = BImagePadLines(bimg);
	np = BImageNextPixel(bimg);

        // Factor for converting current-scale dimensions to
        // real size dimensions
        tile_scale = ((unsigned int) 1) << (pipeData->m_currentMip);

        // Info concerning (whole) image size
        y_maxsize = pipeData->m_mipZeroSize.h;
        x_maxsize = pipeData->m_mipZeroSize.w;

        // Info concerning position of the current tile
        y_offset = bibbleData->m_btf.m_y(pipeData->m_tile);
        x_offset = bibbleData->m_btf.m_x(pipeData->m_tile);

        // Info concerning crop-size
        double cropleft,croptop,cropheight,cropwidth,cropdpi;
        bool cropon;
	BStringHandle optname = BStringAllocFromAscii("cropleft");
	int external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          external_option, 0, cropleft);
	BStringFree(optname);
	optname = BStringAllocFromAscii("croptop");
	external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          external_option, 0, croptop);
	BStringFree(optname);
	optname = BStringAllocFromAscii("cropheight");
	external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          external_option, 0, cropheight);
	BStringFree(optname);
	optname = BStringAllocFromAscii("cropwidth");
	external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          external_option, 0, cropwidth);
	BStringFree(optname);
	optname = BStringAllocFromAscii("cropdpi");
	external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          external_option, 0, cropdpi);
	BStringFree(optname);
	optname = BStringAllocFromAscii("cropon");
	external_option = BGetOptionIDByName(optname, 0 );
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options,
          external_option, 0, cropon);
	BStringFree(optname);
        if(cropdpi<0.0) cropdpi = -cropdpi;
        if(cropon) {
          cropleft *= x_maxsize;
          croptop *= y_maxsize;
          cropheight *= cropdpi;
          cropwidth *= cropdpi;
        } else {
          cropleft = 0.0;
          croptop = 0.0;
          cropheight = y_maxsize;
          cropwidth = x_maxsize;
        }

        fprintf(stdout,"%f %f %f %f (%f)\n",cropleft,croptop,cropwidth,cropheight,cropdpi);

	if ( nChan != 3 )
	{
		BMSG("MainLoopConvert() - Not a 3 channel image\n");
		// We only handle RGB data
		return -1;
	}
        
        // TUTORIAL: grab our settings from the widget
        int frame_width,frame_radius;
        double threshold;
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          SF_WIDTH, bibbleData->m_plugin, frame_width);
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          SF_RADIUS, bibbleData->m_plugin, frame_radius);
	bibbleData->m_bof.m_getOptionDouble( pipeData->m_options,
          SF_THRESHOLD, bibbleData->m_plugin, threshold);
        int rgb;
	bibbleData->m_bof.m_getOptionInt( pipeData->m_options,
          SF_COLOR, bibbleData->m_plugin, rgb);
	double colr = ((double)SfGetRed((unsigned int)rgb))*((double) 257.0);
	double colg = ((double)SfGetGreen((unsigned int)rgb))*((double) 257.0);
	double colb = ((double)SfGetBlue((unsigned int)rgb))*((double) 257.0);

	// Convert each scan line including the padding to monochrome
	// May not be necessary to convert the pad, but any processing
        // after us that dips into the pad
	// may cause color artifacts around the edges of the image
	unsigned short *pIn, *pOut, *pEnd, *pAlpha;
        double R,G,B;

	if ( pipeData->m_layerNum == 0 )
	{
		for( int y = -padh; y < h+padh; ++y ) {
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;
                        unsigned short x = -padw-1;
			while( pIn < pEnd ) {
				R = (double)*(pIn); 
				G = (double)*(pIn + pw); 
				B = (double)*(pIn + 2*pw);
				pIn += np; x++;

// TUTORIAL: write your code here.
// You will have to paste it a second time below.
// This first place will be run on the main layer.
// The other place is for being run on other layers.
// Remember that at this point, R, G, B are the red, green and blue
// value of a given pixel stored as "double" values from 0 to 65535.
// If you need some [0-255] value, remember to divide them by 257 first
// (since 65535/257 = 255).
// If you need some [0-1] value, remember to divide them by 65535.

  unsigned int current_y = (y_offset + y) * tile_scale;
  unsigned int current_x = (x_offset + x) * tile_scale;

  double d = 0;

  if(current_y-croptop<frame_width) d = croptop+frame_width-current_y;
  if(croptop+cropheight-current_y-1<frame_width)
    d = frame_width-croptop-cropheight+current_y+1;
  if(current_x-cropleft<frame_width) d = frame_width+cropleft-current_x;
  if(cropleft+cropwidth-current_x-1<frame_width)
    d = frame_width-cropleft-cropwidth+current_x+1;

  if((current_y-croptop<frame_width+frame_radius)&&
     (current_x-cropleft<frame_width+frame_radius)) {
       double cx = cropleft+frame_width+frame_radius;
       double cy = croptop+frame_width+frame_radius;
       d = sqrt(
         (current_y-cy)*(current_y-cy) +
         (current_x-cx)*(current_x-cx)
         );
       if(d<frame_radius) d = 0.0;
       else d -= frame_radius;
  }
  if((croptop+cropheight-current_y-1<frame_width+frame_radius)&&
     (current_x-cropleft<frame_width+frame_radius)) {
       double cx = cropleft+frame_width+frame_radius;
       double cy = croptop+cropheight-frame_width-1-frame_radius;
       d = sqrt(
         (current_y-cy)*(current_y-cy) +
         (current_x-cx)*(current_x-cx)
         );
       if(d<frame_radius) d = 0.0;
       else d -= frame_radius;
  }
  if((current_y-croptop<frame_width+frame_radius)&&
     (cropleft+cropwidth-current_x-1<frame_width+frame_radius)) {
       double cx = cropleft+cropwidth-frame_width-1-frame_radius;
       double cy = croptop+frame_width+frame_radius;
       d = sqrt(
         (current_y-cy)*(current_y-cy) +
         (current_x-cx)*(current_x-cx)
         );
       if(d<frame_radius) d = 0.0;
       else d -= frame_radius;
  }
  if((croptop+cropheight-current_y-1<frame_width+frame_radius)&&
     (cropleft+cropwidth-current_x-1<frame_width+frame_radius)) {
       double cx = cropleft+cropwidth-frame_width-1-frame_radius;
       double cy = croptop+cropheight-frame_width-1-frame_radius;
       d = sqrt(
         (current_y-cy)*(current_y-cy) +
         (current_x-cx)*(current_x-cx)
         );
       if(d<frame_radius) d = 0.0;
       else d -= frame_radius;
  }
  d /= 256.0;
  if(d>0.0) {
    if(d>threshold) d = 1.0;
    else {
      if(threshold<0.005) d = 1.0;
      else d /= threshold;
    }
  }

  R = R*(1.0-d) + d*colr;
  G = G*(1.0-d) + d*colg;
  B = B*(1.0-d) + d*colb;

// TUTORIAL: end of your code here.
// Don't forget to copy it below.

                                if(R<0.0f) R=0.0f;
                                if(R>65535.0f) R=65535.0f;
                                if(G<0.0f) G=0.0f;
                                if(G>65535.0f)
                                G=65535.0f;
                                if(B<0.0f) B=0.0f;
                                if(B>65535.0f) B=65535.0f;

				*(pOut) = (unsigned short)(R + 0.5f);
				*(pOut + pw) = (unsigned short)(G + 0.5f);
				*(pOut + 2*pw) = (unsigned short)(B + 0.5f);
				pOut += np;
			}
		}
	} else {
		float ival, aval;
		unsigned int oval;
		for( int y = -padh; y < h+padh; ++y ) 
		{
			pIn = BImageScanLine16( bimg, 0 ) + y*rw - padw*np;
			pOut = BImageScanLine16( dest, 0 ) + y*rw - padw*np;
			pAlpha = BImageScanLine16( alpha, 0 ) + y*rw - padw*np;
			pEnd = pIn + (w+padw*2)*np;

                        unsigned short x = -padw-1;
			while( pIn < pEnd ) {
				aval = *pAlpha;
				ival = 65535 - aval;

				if ( aval == 0 )
				{
					// Completely masked out, just skip this pixel
					pIn += np; pOut += np; pAlpha += np;
                                        x++;
					continue;
				}

				R = (double)*(pIn); 
				G = (double)*(pIn + pw); 
				B = (double)*(pIn + 2*pw);
// TUTORIAL: copy your code here.
// You probably already put it above.
// This first place will be run on the main layer.
// This new place is for being run on other layers.
// Remember that at this point, R, G, B are the red, green and blue
// value of a given pixel stored as "double" values from 0 to 65535.
// If you need some [0-255] value, remember to divide them by 257 first
// (since 65535/257 = 255).
// If you need some [0-1] value, remember to divide them by 65535.

// TUTORIAL: end of your code here.
// Don't forget to copy it below.

				if ( ival == 0 )
				{
					// Completely transparent, just copy
					*(pOut) = (unsigned short)(R + 0.5f);
					*(pOut + pw) = (unsigned short)(G + 0.5f);
					*(pOut + 2*pw) = (unsigned short)(B + 0.5f);
				} else {
					oval = pOut[0]*ival + R*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[0] = oval;

					oval = pOut[pw]*ival + G*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;

					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw] = oval;

					oval = pOut[pw+pw]*ival + B*aval + ((double) .5);
					oval >>= 16;
					oval = oval & 0x0000ffff;
					if (oval < 0)
						oval = 0;
					if (oval > 65535)
						oval = 65535;
					pOut[pw+pw] = oval;

				}
				pOut += np; pIn += np; pAlpha += np; x++;
			}
		}
	}
	return rc;
}

// You may name the function anything you like, but if you want to use the macros to
// call the Bibble functions, you must keep the parameter names the same
// typedef int	(*OptionsInitPtr )(struct BibblePluginData *bibbleData, struct OptionData *od);

int initializeOptions(struct  BibblePluginData *bibbleData, BSettingsWidgetHandle sw)
{
	// This function is called when the plugin loads
	int rc = 0;

	return rc;
}

/*
  The following function is called in order to fix dependancies between
  settings (for instance, if some user has to choose a color with three
  sliders (red, green, blue), you may want to change sliders for some
  Hue, saturation, value model of the same color (HSV) in order to
  allow the user to choose a color by two different (but linked) ways.
  You may see some dependancies with the Simple Color Filter v1.1 (just
  set the color with the sliders, and see the color of the button changing
  in the same time).
  The most known feature handled here is enabling the plugin when the user
  moves a slider.
*/
int doDependancies( struct BibblePluginData *bibbleData, struct OptionData *od) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Simple Frame Plugin - Do Dependancies");
#endif

        // A nice feature: when the user will change some settings,
        // the plugin will get enabled "automagically"
        // The double-bar || means "or": if there is a change on
        // some option OR on some other option, then enable the plugin.
	if ((BOptionListContains(od->m_changeList,SF_WIDTH,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SF_RADIUS,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SF_THRESHOLD,bibbleData->m_plugin)) || (BOptionListContains(od->m_changeList,SF_COLOR,bibbleData->m_plugin)))
	{
		BSetOptionBool(od->m_changeList,SF_ENABLED,bibbleData->m_plugin,true);
	}

	return rc;
}

// This function is a registered button callback
int sfHandleButton( struct BibblePluginData *bibbleData, struct OptionData *od ) {
	int rc = 0;

#ifdef _DEBUG
	BMSG("Simple Frame Plugin - sfHandleButton()" );
#endif

	BMsgBox( "Simple Frame Plugin", "Version 1.0\n\nWritten by Thomas Baruchel\nhttp://baruchel.free.fr/bibble/\n\nThis plugin is an open-source software released under\nthe terms of the MIT license.\nhttp://www.opensource.org/licenses/mit-license.php\n\nIt is not only free as in beer, but also free as in freedom.\nYou may use it, you may share it, you may change it, you\nmay even sell it." );
	return rc;
}

__declspec(dllexport) int RegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble during loading
	// Plug-ins should register the functions Bibble needs to call,
	// check that they are compatible with the version of Bibble attempting to load them,
	// add their settings to the factory defaults list for the Bibble option system,
	// and allocate and initialize any global data or resources
	int rc = 0;

	BTestDebug = (BDebugTraceFunction)(bibbleData->m_buf.m_trace);

	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	// Set the API version we were compiled under
	// This is defined in BibblePlugin.h
	bibbleData->m_version = BIBBLE5PLUGIN_API_VERSION;
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_groupName, SF_PLUGINGROUP);
	bibbleData->m_bsf.m_fromAscii( bibbleData->m_pluginName, SF_PLUGINNAME);

        // TUTORIAL: choose the stage where you want the plugin
        // to run (you may find the list in the BibblePlugin.h
        // file from the SDK); you may also choose some priority
        // (in order tomake it run before or after another one
        // in the same stage): high priority means that the
        // plugin will run earlier.
        // In order to make some tests, see if your plugin runs
        // before or after the Black and White plugin which has
        // the following settings: PIPE_PLUGIN_CURVES, 200
        // Try a lower and then a higher priority and look at
        // the result.
	bibbleData->m_brf.m_registerPipeRoutines(bibbleData->m_plugin, MainLoopConvert, PIPE_PLUGIN_FINAL, 200);
	bibbleData->m_brf.m_registerOptionsInit(bibbleData->m_plugin, initializeOptions);

	// Register my button callback
	bibbleData->m_brf.m_registerButtonCallback(bibbleData->m_plugin, sfHandleButton, SF_BUTTONNAME);

        // The function registered below (doDependancies) will be called
        // when a setting will be changed
	bibbleData->m_brf.m_registerSettingsWidgetDoDependancies(bibbleData->m_plugin, doDependancies );

	BMSG("Simple Frame Plugin plugin loaded");
	return rc;
}

__declspec(dllexport) int unRegisterPlugin(struct BibblePluginData *bibbleData)
{
	// This function is called by Bibble right before the program exits
	// Plug-ins should free any global memory and resources here
	int rc = 0;

	return rc;
}

__declspec(dllexport) int getVersion( void )
{
	return BIBBLE5PLUGIN_API_VERSION;
}

/*
  TUTORIAL:
  The following function is very easy to handle but you have to
  spend a little time on it. Add one line for each setting by
  copying or changing existing lines.
	baf->m_addIntOption    (for an integer value) 
	baf->m_addDoubleOption (for a double value)
	baf->m_addBoolOption   (for a boolean value, most often checkboxes)
  The third parameter is the name of the setting in the UI (without the
  suffix: "_slider", "_edit", "_checkbox", etc.
  The fourth parameter is the name of the setting in the "selective settings"
  panel (something human-readable but rather short: three or four words).
  The sixth parameter is the tooltip for this setting (but don't forget to
  copy the same tooltip in the UI file).
  The seventh parameter is the default value for this setting (but don't
  forget to put this default value in the UI which can be relevant in some
  situations).
  The eightth parameter tells Bibble how to handle this setting. You may choose
  several options (all are in the Bibble.h file from the SDK); if you want
  to use several, use a single vertical bar | between them (you may see
  an example in the first line below).
  Most often you can only put OPTION_RESTART which tells that Bibble has to
  restart the computation for this picture when the setting has changed.
  For the "enabled" checkbox I also added a second option to tell Bibble
  it does not need to check dependancies (see the doDependancies function
  above) when the user changes this setting: no other setting has to be
  linked to this action.
*/
__declspec(dllexport) int buildDefaults(struct BibblePluginData *bibbleData, struct BAddOptionFunctions *baf)
{
	int rc = 0;

	baf->m_addBoolOption( bibbleData->m_plugin, SF_ENABLED, "sf_enabled", "Enable", SF_GROUP, "Enable Simple Frame Plugin.", false, OPTION_RESTART|OPTION_NO_DEPENDENCY);
	baf->m_addIntOption( bibbleData->m_plugin, SF_WIDTH, "sf_width", "Width", SF_GROUP, "Set the width of the frame.", 128, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin, SF_RADIUS, "sf_radius", "Radius", SF_GROUP, "Set the radius at the corners.", 128, OPTION_RESTART);
	baf->m_addDoubleOption( bibbleData->m_plugin, SF_THRESHOLD, "sf_threshold", "Feathering", SF_GROUP, "Apply some feathering effect.", 0.5, OPTION_RESTART);
	baf->m_addIntOption( bibbleData->m_plugin,SF_COLOR, "sf_color", "Color", SF_GROUP, "Set the color of the frame.", -1, OPTION_RESTART);
	return rc;
}

__declspec(dllexport) bool runCheck(struct BibblePluginData *bibbleData, struct BPluginPipeData *pipeData, BBOOL &needPrev )
{
	bool enabled = false;
	needPrev = false;
	bibbleData->m_bof.m_getOptionBool( pipeData->m_options, SF_ENABLED, bibbleData->m_plugin, enabled);
	return enabled;
}
